"""Custom type aliases.

This module defines commonly used types in the library. These are separated into two
different kinds, API types and argument types.

API types are aliases which define custom types used throughout the library. Objects of
this type may be supplied as arguments or returned by a method.

Argument types are aliases which define commonly used method arguments. These should
only ever be used in the signature of a method and then be converted internally, e.g.
in a class instantiation or an interface. They enable the user to conveniently
specify a variety of object types for the same argument, while ensuring a unified
internal representation of those same objects.
"""

import numbers
from typing import Iterable, Tuple, Union

import numpy as np

########################################################################################
# API Types
########################################################################################

ShapeType = Tuple[int, ...]

RandomStateType = Union[np.random.RandomState, np.random.Generator]
"""Type of a random number generator."""

########################################################################################
# Argument Types
########################################################################################

IntArgType = Union[int, numbers.Integral, np.integer]
FloatArgType = Union[float, numbers.Real, np.floating]

ShapeArgType = Union[IntArgType, Iterable[IntArgType]]
"""Type of a public API argument for supplying a shape. Values of this type should
always be converted into :class:`ShapeType` using the function
:func:`probnum.utils.as_shape` before further internal processing."""

DTypeArgType = Union[np.dtype, str]
"""Type of a public API argument for supplying a dtype. Values of this type should
always be converted into :class:`np.dtype` using the function
:func:`np.dtype` before further internal processing."""

ScalarArgType = Union[int, float, complex, numbers.Number, np.number]
"""Type of a public API argument for supplying a scalar value. Values of this type
should always be converted into :class:`np.generic` using the function
:func:`probnum.utils.as_scalar` before further internal processing."""

ArrayLikeGetitemArgType = Union[
    int,
    slice,
    np.ndarray,
    np.newaxis,
    None,
    type(Ellipsis),
    Tuple[Union[int, slice, np.ndarray, np.newaxis, None, type(Ellipsis)], ...],
]

RandomStateArgType = Union[None, int, np.random.RandomState, np.random.Generator]
"""Type of a public API argument for supplying a random number generator. Values of this
type should always be converted into :class:`RandomStateType` using the function
:func:`probnum.utils.as_random_state` before further internal processing."""



import numpy as np

from probnum.problems import InitialValueProblem

__all__ = ["threebody_jax", "vanderpol_jax"]
# pylint: disable=import-outside-toplevel


JAX_ERRORMSG = "IVP instantiation requires jax. Try using the pure numpy versions instead, or install `jax` via `pip install jax jaxlib`"


def threebody_jax(tmax=17.0652165601579625588917206249):
    r"""Initial value problem (IVP) based on a three-body problem.

    Let the initial conditions be :math:`y = (y_1, y_2, \dot{y}_1, \dot{y}_2)^T`.
    This function implements the second-order three-body problem as a system of
    first-order ODEs, which is defined as follows: [1]_

    .. math::

        f(t, y) =
        \begin{pmatrix}
            \dot{y_1} \\
            \dot{y_2} \\
            y_1 + 2 \dot{y}_2 - \frac{(1 - \mu) (y_1 + \mu)}{d_1}
                - \frac{\mu (y_1 - (1 - \mu))}{d_2} \\
            y_2 - 2 \dot{y}_1 - \frac{(1 - \mu) y_2}{d_1} - \frac{\mu y_2}{d_2}
        \end{pmatrix}

    with

    .. math::

        d_1 &= ((y_1 + \mu)^2 + y_2^2)^{\frac{3}{2}} \\
        d_2 &= ((y_1 - (1 - \mu))^2 + y_2^2)^{\frac{3}{2}}

    and a constant parameter  :math:`\mu = 0.012277471` denoting the standardized moon mass.

    Parameters
    ----------
    tmax
        Final time.

    Returns
    -------
    IVP
        IVP object describing a three-body problem IVP with the prescribed
        configuration.

    References
    ----------
    .. [1] Hairer, E., Norsett, S. and Wanner, G..
        Solving Ordinary Differential Equations I.
        Springer Series in Computational Mathematics, 1993.
    """

    try:
        import jax
        import jax.numpy as jnp
        from jax.config import config

        config.update("jax_enable_x64", True)

    except ImportError as err:
        raise ImportError(JAX_ERRORMSG) from err

    def threebody_rhs(Y):
        # defining the ODE:
        # assume Y = [y1,y2,y1',y2']
        mu = 0.012277471  # a constant (standardised moon mass)
        mp = 1 - mu
        D1 = ((Y[0] + mu) ** 2 + Y[1] ** 2) ** (3 / 2)
        D2 = ((Y[0] - mp) ** 2 + Y[1] ** 2) ** (3 / 2)
        y1p = Y[0] + 2 * Y[3] - mp * (Y[0] + mu) / D1 - mu * (Y[0] - mp) / D2
        y2p = Y[1] - 2 * Y[2] - mp * Y[1] / D1 - mu * Y[1] / D2
        return jnp.array([Y[2], Y[3], y1p, y2p])

    df = jax.jacfwd(threebody_rhs)
    ddf = jax.jacrev(df)

    def rhs(t, y):
        return threebody_rhs(Y=y)

    def jac(t, y):
        return df(y)

    def hess(t, y):
        return ddf(y)

    y0 = np.array([0.994, 0, 0, -2.00158510637908252240537862224])
    t0 = 0.0
    return InitialValueProblem(f=rhs, t0=t0, tmax=tmax, y0=y0, df=jac, ddf=hess)


def vanderpol_jax(t0=0.0, tmax=30, y0=None, params=1e1):
    r"""Initial value problem (IVP) based on the Van der Pol Oscillator, implemented in `jax`.

    This function implements the second-order Van-der-Pol Oscillator as a system
    of first-order ODEs.
    The Van der Pol Oscillator is defined as

    .. math::

        f(t, y) =
        \begin{pmatrix}
            y_2 \\
            \mu \cdot (1 - y_1^2)y_2 - y_1
        \end{pmatrix}

    for a constant parameter  :math:`\mu`.
    :math:`\mu` determines the stiffness of the problem, where
    the larger :math:`\mu` is chosen, the more stiff the problem becomes.
    Default is :math:`\mu = 0.1`.
    This implementation includes the Jacobian :math:`J_f` of :math:`f`.

    Parameters
    ----------
    t0 : float
        Initial time point. Leftmost point of the integration domain.
    tmax : float
        Final time point. Rightmost point of the integration domain.
    y0 : np.ndarray,
        *(shape=(2, ))* -- Initial value of the problem.
    params : (float), optional
        Parameter :math:`\mu` for the Van der Pol Equations
        Default is :math:`\mu=0.1`.

    Returns
    -------
    IVP
        IVP object describing the Van der Pol Oscillator IVP with the prescribed
        configuration.
    """

    try:
        import jax
        import jax.numpy as jnp
        from jax.config import config

        config.update("jax_enable_x64", True)

    except ImportError as err:
        raise ImportError(JAX_ERRORMSG) from err

    if isinstance(params, float):
        mu = params
    else:
        (mu,) = params

    if y0 is None:
        y0 = np.array([2.0, 0.0])

    def vanderpol_rhs(Y):
        return jnp.array([Y[1], mu * (1.0 - Y[0] ** 2) * Y[1] - Y[0]])

    df = jax.jacfwd(vanderpol_rhs)
    ddf = jax.jacrev(df)

    def rhs(t, y):
        return vanderpol_rhs(Y=y)

    def jac(t, y):
        return df(y)

    def hess(t, y):
        return ddf(y)

    return InitialValueProblem(f=rhs, t0=t0, tmax=tmax, y0=y0, df=jac, ddf=hess)



"""Benchmarks for probabilistic IVP solvers."""
import numpy as np

from probnum.diffeq import lotkavolterra, probsolve_ivp
from probnum.randvars import Constant


def load_lotkavolterra():
    """Load LV system as a basic IVP."""
    initrv = Constant(np.array([20, 20]))
    return lotkavolterra(
        timespan=[0, 0.55], initrv=initrv, params=(0.5, 0.05, 0.5, 0.05)
    )


class IVPSolve:
    """Benchmark ODE-filter and ODE-smoother with fixed steps and different priors and
    information operators."""

    param_names = ["method", "algo_order"]
    params = [["ek0", "ek1"], [2, 3]]

    def setup(self, method, prior):
        # pylint: disable=invalid-name
        self.ivp = load_lotkavolterra()
        self.stepsize = 1e-1

    def time_solve(self, method, algo_order):
        f = self.ivp.rhs
        df = self.ivp.jacobian
        t0, tmax = self.ivp.timespan
        y0 = self.ivp.initrv.mean
        probsolve_ivp(
            f,
            t0,
            tmax,
            y0,
            df=df,
            method=method,
            dense_output=True,
            algo_order=algo_order,
            step=self.stepsize,
            adaptive=False,
        )

    def peakmem_solve(self, method, algo_order):
        f = self.ivp.rhs
        df = self.ivp.jacobian
        t0, tmax = self.ivp.timespan
        y0 = self.ivp.initrv.mean
        probsolve_ivp(
            f,
            t0,
            tmax,
            y0,
            df=df,
            method=method,
            dense_output=True,
            algo_order=algo_order,
            step=self.stepsize,
            adaptive=False,
        )



"""Test fixtures for kernels."""

from typing import Optional

import numpy as np
import pytest

import probnum.kernels as kernels


@pytest.fixture(
    params=[pytest.param(seed, id=f"seed{seed}") for seed in range(1)],
    name="random_state",
)
def fixture_random_state(request):
    """Random state(s) used for test parameterization."""
    return np.random.RandomState(seed=request.param)


@pytest.fixture(
    params=[
        pytest.param(num_data, id=f"ndata{num_data}") for num_data in [1, 2, 10, 100]
    ],
    name="num_data",
)
def fixture_num_data(request) -> int:
    """Size of the dataset."""
    return request.param


@pytest.fixture(
    params=[
        pytest.param(input_dim, id=f"indim{input_dim}") for input_dim in [1, 10, 100]
    ],
    name="input_dim",
)
def fixture_input_dim(request) -> int:
    """Input dimension of the covariance function."""
    return request.param


@pytest.fixture(
    params=[
        pytest.param(output_dim, id=f"outdim{output_dim}") for output_dim in [1, 2, 10]
    ]
)
def output_dim(request) -> int:
    """Output dimension of the covariance function."""
    return request.param


# Datasets
@pytest.fixture(name="x0")
def fixture_x0(
    num_data: int, input_dim: int, random_state: np.random.RandomState
) -> np.ndarray:
    """Random data from a standard normal distribution."""
    return random_state.normal(0, 1, size=(num_data, input_dim))


@pytest.fixture(
    params=[
        pytest.param(num_data, id=f"ndata{num_data}") for num_data in [None, 10, 2, 1]
    ],
    name="x1",
)
def fixture_x1(
    request, input_dim: int, random_state: np.random.RandomState
) -> Optional[np.ndarray]:
    """Random data from a standard normal distribution."""
    if request.param is None:
        return None
    else:
        return random_state.normal(0, 1, size=(request.param, input_dim))


@pytest.fixture()
def x0_1d(input_dim: int, random_state: np.random.RandomState) -> np.ndarray:
    """Random 1D dataset."""
    return random_state.normal(0, 1, size=(input_dim,))


# Kernel and kernel matrices


@pytest.fixture(
    params=[
        pytest.param(kerndef, id=kerndef[0].__name__)
        for kerndef in [
            (kernels.Linear, {"constant": 1.0}),
            (kernels.WhiteNoise, {"sigma": -1.0}),
            (kernels.Polynomial, {"constant": 1.0, "exponent": 3}),
            (kernels.ExpQuad, {"lengthscale": 1.5}),
            (kernels.RatQuad, {"lengthscale": 0.5, "alpha": 2.0}),
            (kernels.Matern, {"lengthscale": 0.5, "nu": 0.5}),
            (kernels.Matern, {"lengthscale": 0.5, "nu": 1.5}),
            (kernels.Matern, {"lengthscale": 1.5, "nu": 2.5}),
            (kernels.Matern, {"lengthscale": 2.5, "nu": 7.0}),
            (kernels.Matern, {"lengthscale": 3.0, "nu": np.inf}),
        ]
    ],
    name="kernel",
)
def fixture_kernel(request, input_dim: int) -> kernels.Kernel:
    """Kernel / covariance function."""
    return request.param[0](**request.param[1], input_dim=input_dim)


@pytest.fixture()
def kernmat(
    kernel: kernels.Kernel, x0: np.ndarray, x1: Optional[np.ndarray]
) -> np.ndarray:
    """Kernel evaluated at the data."""
    return kernel(x0=x0, x1=x1)



import unittest

import numpy as np

from probnum.diffeq import ConstantSteps, ODESolution, ODESolver, logistic
from probnum.randvars import Constant


class MockODESolver(ODESolver):
    """Euler method as an ODE solver."""

    def initialise(self):
        return self.ivp.t0, self.ivp.initrv

    def step(self, start, stop, current):
        h = stop - start
        x = current.mean
        xnew = x + h * self.ivp(start, x)
        return (
            Constant(xnew),
            np.nan,
        )  # return nan as error estimate to ensure that it is not used

    def rvlist_to_odesol(self, times, rvs):
        return ODESolution(locations=times, states=rvs)


class ODESolverTestCase(unittest.TestCase):
    """An ODE Solver has to work with just step() and initialise() provided.

    We implement Euler in MockODESolver to assure this.
    """

    def setUp(self):
        y0 = Constant(0.3)
        ivp = logistic([0, 4], initrv=y0)
        euler_order = 1
        self.solver = MockODESolver(ivp, order=euler_order)
        self.step = 0.2

    def test_solve(self):
        steprule = ConstantSteps(self.step)
        odesol = self.solver.solve(
            steprule=steprule,
        )  # this is the actual part of the test

        # quick check that the result is sensible
        self.assertAlmostEqual(odesol.locations[-1], self.solver.ivp.tmax)
        self.assertAlmostEqual(odesol.states[-1].mean, 1.0, places=2)


if __name__ == "__main__":
    unittest.main()



"""Example Implementation of 1D (noisy) quadratic optimization."""

from ._probsolve_qp import probsolve_qp
from .probabilistic_quadratic_optimizer import ProbabilisticQuadraticOptimizer

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "probsolve_qp",
    "ProbabilisticQuadraticOptimizer",
]

# Set correct module paths. Corrects links and module paths in documentation.
ProbabilisticQuadraticOptimizer.__module__ = "probnum.quadopt_example"



"""Contains integration measures."""

import abc
from typing import Optional, Tuple, Union

import numpy as np
import scipy.stats

from probnum.randvars import Normal
from probnum.type import FloatArgType, IntArgType, RandomStateArgType


class IntegrationMeasure(abc.ABC):
    """An abstract class for a measure against which a target function is integrated.

    Child classes implement specific integration measures and, if available, make use
    of random variables for sampling and evaluation of the density function.

    Parameters
    ----------
    dim :
        Dimension of the integration domain.
    domain :
        Tuple which contains two arrays which define the start and end points,
        respectively, of the rectangular integration domain.
    """

    def __init__(
        self,
        dim: IntArgType,
        domain: Tuple[Union[np.ndarray, FloatArgType], Union[np.ndarray, FloatArgType]],
    ) -> None:

        self._set_dimension_domain(dim, domain)

    def __call__(self, points: Union[float, np.floating, np.ndarray]) -> np.ndarray:
        """Evaluate the density function of the integration measure.

        Parameters
        ----------
        points :
            *shape=(n_points,) or (n_points, dim)* -- Input locations.

        Returns
        -------
        density_evals :
            *shape=(n_points,)* -- Density evaluated at given locations.
        """
        # pylint: disable=no-member
        return self.random_variable.pdf(points).squeeze()

    def sample(
        self,
        n_sample: IntArgType,
        random_state: Optional[RandomStateArgType] = None,
    ) -> np.ndarray:
        """Sample ``n_sample`` points from the integration measure.

        Parameters
        ----------
        n_sample :
            Number of points to be sampled
        random_state :
            Random state of the random variable corresponding to the integration measure.

        Returns
        -------
        points :
            *shape=(n_sample,) or (n_sample,dim)* -- Sampled points
        """
        # pylint: disable=no-member
        self.random_variable.random_state = random_state
        return np.reshape(
            self.random_variable.sample(size=n_sample), newshape=(n_sample, self.dim)
        )

    def _set_dimension_domain(
        self,
        dim: IntArgType,
        domain: Tuple[Union[np.ndarray, FloatArgType], Union[np.ndarray, FloatArgType]],
    ) -> None:
        """Sets the integration domain and dimension.

        The following logic is used to set the domain and dimension:
            1. If ``dim`` is not given (``dim == None``):
                1a. If either ``domain[0]`` or ``domain[1]`` is a scalar, the dimension
                    is set as the maximum of their lengths and the scalar is expanded to
                    a constant vector.
                1b. Otherwise, if the ``domain[0]`` and ``domain[1]`` are not of equal
                    length, an error is raised.
            2. If ``dim`` is given:
                2a. If both ``domain[0]`` and ``domain[1]`` are scalars, they are
                    expanded to constant vectors of length ``dim``.
                2b. If only one of `domain[0]`` and ``domain[1]`` is a scalar and the
                    length of the other equals ``dim``, the scalar one is expanded to a
                    constant vector of length ``dim``.
                2c. Otherwise, if neither of ``domain[0]`` and ``domain[1]`` is a
                    scalar, error is raised if either of them has length which does not
                    equal ``dim``.
        """
        domain_a_dim = np.size(domain[0])
        domain_b_dim = np.size(domain[1])

        # Check that given dimensions match and are positive
        dim_mismatch = False
        if dim is None:
            if domain_a_dim == domain_b_dim:
                dim = domain_a_dim
            elif domain_a_dim == 1 or domain_b_dim == 1:
                dim = np.max([domain_a_dim, domain_b_dim])
            else:
                dim_mismatch = True
        else:
            if (domain_a_dim > 1 or domain_b_dim > 1) and dim != np.max(
                [domain_a_dim, domain_b_dim]
            ):
                dim_mismatch = True

        if dim_mismatch:
            raise ValueError(
                "Domain limits must have the same length or at least "
                "one of them has to be one-dimensional."
            )
        if dim < 1:
            raise ValueError(f"Domain dimension dim = {dim} must be positive.")

        # Use same domain limit in all dimensions if only one limit is given
        if domain_a_dim == 1:
            domain_a = np.full((dim,), domain[0])
        else:
            domain_a = domain[0]
        if domain_b_dim == 1:
            domain_b = np.full((dim,), domain[1])
        else:
            domain_b = domain[1]

        # Check that the domain is non-empty
        if not np.all(domain_a < domain_b):
            raise ValueError("Domain must be non-empty.")

        self.dim = dim
        self.domain = (domain_a, domain_b)


class LebesgueMeasure(IntegrationMeasure):
    """Lebesgue measure on a hyper-rectangle.

    Parameters
    ----------
    dim :
        Dimension of the integration domain
    domain :
        Tuple which contains two arrays which define the start and end points,
        respectively, of the rectangular integration domain.
    normalized :
         Boolean which controls whether or not the measure is normalized (i.e.,
         integral over the domain is one).
    """

    def __init__(
        self,
        domain: Tuple[Union[np.ndarray, FloatArgType], Union[np.ndarray, FloatArgType]],
        dim: Optional[IntArgType] = None,
        normalized: Optional[bool] = False,
    ) -> None:
        super().__init__(dim=dim, domain=domain)

        # Set normalization constant
        self.normalized = normalized
        if self.normalized:
            self.normalization_constant = 1.0 / np.prod(self.domain[1] - self.domain[0])
        else:
            self.normalization_constant = 1.0

        if self.normalization_constant in [0, np.Inf, -np.Inf]:
            raise ValueError(
                "Normalization constant is too small or too large. "
                "Consider setting normalized = False."
            )

        # Use scipy's uniform random variable since uniform random variables are not
        # yet implemented in probnum
        self.random_variable = scipy.stats.uniform(
            loc=self.domain[0], scale=self.domain[1] - self.domain[0]
        )

    def __call__(self, points: Union[float, np.floating, np.ndarray]) -> np.ndarray:
        num_dat = np.atleast_1d(points).shape[0]
        return np.full(() if num_dat == 1 else (num_dat,), self.normalization_constant)

    def sample(
        self,
        n_sample: IntArgType,
        random_state: Optional[RandomStateArgType] = None,
    ) -> np.ndarray:
        return self.random_variable.rvs(
            size=(n_sample, self.dim), random_state=random_state
        )


# pylint: disable=too-few-public-methods
class GaussianMeasure(IntegrationMeasure):
    """Gaussian measure on Euclidean space with given mean and covariance.

    If ``mean`` and ``cov`` are scalars but ``dim`` is larger than one, ``mean`` and
    ``cov`` are extended to a constant vector and diagonal matrix, respectively,
    of appropriate dimensions.

    Parameters
    ----------
    mean :
        *shape=(dim,)* -- Mean of the Gaussian measure.
    cov :
        *shape=(dim, dim)* -- Covariance matrix of the Gaussian measure.
    dim :
        Dimension of the integration domain.
    """

    def __init__(
        self,
        mean: Union[float, np.floating, np.ndarray],
        cov: Union[float, np.floating, np.ndarray],
        dim: Optional[IntArgType] = None,
    ) -> None:

        # Extend scalar mean and covariance to higher dimensions if dim has been
        # supplied by the user
        # pylint: disable=fixme
        # TODO: This needs to be modified to account for cases where only either the
        #  mean or covariance is given in scalar form
        if (
            (np.isscalar(mean) or mean.size == 1)
            and (np.isscalar(cov) or cov.size == 1)
            and dim is not None
        ):
            mean = np.full((dim,), mean)
            cov = cov * np.eye(dim)

        # Set dimension based on the mean vector
        if np.isscalar(mean):
            dim = 1
        else:
            dim = mean.size

        # If cov has been given as a vector of variances, transform to diagonal matrix
        if isinstance(cov, np.ndarray) and np.squeeze(cov).ndim == 1 and dim > 1:
            cov = np.diag(np.squeeze(cov))

        # Exploit random variables to carry out mean and covariance checks
        self.random_variable = Normal(mean=np.squeeze(mean), cov=np.squeeze(cov))
        self.mean = self.random_variable.mean
        self.cov = self.random_variable.cov

        # Set diagonal_covariance flag
        if dim == 1:
            self.diagonal_covariance = True
        else:
            self.diagonal_covariance = (
                np.count_nonzero(self.cov - np.diag(np.diagonal(self.cov))) == 0
            )

        super().__init__(
            dim=dim,
            domain=(np.full((dim,), -np.Inf), np.full((dim,), np.Inf)),
        )



"""Tests for Gaussian processes."""

import numpy as np
import pytest

from probnum import randprocs, randvars, utils


def test_no_kernel_covariance_raises_error():
    """Initializing a GP with a covariance function which is not a kernel raises a
    TypeErrror."""
    with pytest.raises(TypeError):
        randprocs.GaussianProcess(mean=np.zeros_like, cov=np.dot)


def test_finite_evaluation_is_normal(gaussian_process: randprocs.GaussianProcess):
    """A Gaussian process evaluated at a finite set of inputs is a Gaussian random
    variable."""
    x = np.random.normal(size=(5,) + utils.as_shape(gaussian_process.input_dim))
    assert isinstance(gaussian_process(x), randvars.Normal)



import functools
import operator
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
import scipy.sparse

import probnum.utils
from probnum.type import ScalarArgType, ShapeArgType

from ._linear_operator import (  # pylint: disable=cyclic-import
    BinaryOperandType,
    LinearOperator,
    Matrix,
)
from ._scaling import Scaling


def add(op1: BinaryOperandType, op2: BinaryOperandType) -> LinearOperator:
    op1, op2 = _operands_to_compatible_linops(op1, op2)

    return _apply(_add_fns, op1, op2, fallback_operator=SumLinearOperator)


def sub(op1: BinaryOperandType, op2: BinaryOperandType) -> LinearOperator:
    op1, op2 = _operands_to_compatible_linops(op1, op2)

    return _apply(
        _sub_fns,
        op1,
        op2,
        fallback_operator=lambda op1, op2: SumLinearOperator(op1, -op2),
    )


def mul(op1: BinaryOperandType, op2: BinaryOperandType) -> LinearOperator:
    return _apply(_mul_fns, op1, op2, fallback_operator=_mul_fallback)


def matmul(op1: LinearOperator, op2: LinearOperator) -> LinearOperator:
    return _apply(_matmul_fns, op1, op2, fallback_operator=_matmul_fallback)


########################################################################################
# Operator registry
########################################################################################

_BinaryOperatorType = Callable[
    [LinearOperator, LinearOperator], Union[LinearOperator, type(NotImplemented)]
]
_BinaryOperatorRegistryType = Dict[Tuple[type, type], _BinaryOperatorType]


_add_fns: _BinaryOperatorRegistryType = {}
_sub_fns: _BinaryOperatorRegistryType = {}
_mul_fns: _BinaryOperatorRegistryType = {}
_matmul_fns: _BinaryOperatorRegistryType = {}


def _apply(
    op_registry: _BinaryOperatorRegistryType,
    op1: LinearOperator,
    op2: LinearOperator,
    fallback_operator: Optional[
        Callable[
            [LinearOperator, LinearOperator],
            Union[LinearOperator, type(NotImplemented)],
        ]
    ] = None,
) -> Union[LinearOperator, type(NotImplemented)]:
    key = (type(op1), type(op2))

    if key in op_registry:
        res = op_registry[key](op1, op2)
    else:
        res = NotImplemented

    if res is NotImplemented and fallback_operator is not None:
        res = fallback_operator(op1, op2)

    return res


########################################################################################
# Helper Functions
########################################################################################


def _operand_to_linop(operand: Any, shape: ShapeArgType) -> Optional[LinearOperator]:
    if isinstance(operand, LinearOperator):
        pass
    elif np.ndim(operand) == 0:
        operand = Scaling(operand, shape=shape)
    elif isinstance(operand, (np.ndarray, scipy.sparse.spmatrix)):
        operand = Matrix(operand)
    else:
        operand = None

    return operand


def _operands_to_compatible_linops(
    op1: Any, op2: Any
) -> Tuple[Optional[LinearOperator], Optional[LinearOperator]]:
    if not isinstance(op1, LinearOperator) and not isinstance(op2, LinearOperator):
        raise TypeError("At least one of the two operands must be a `LinearOperator`.")
    elif not isinstance(op1, LinearOperator):
        op1 = _operand_to_linop(op1, shape=op2.shape)
    elif not isinstance(op2, LinearOperator):
        op2 = _operand_to_linop(op2, shape=op1.shape)
    else:
        # TODO: check whether both have the same shape
        pass

    return op1, op2


########################################################################################
# Generic Linear Operator Arithmetic (Fallbacks)
########################################################################################


class ScaledLinearOperator(LinearOperator):
    """Linear operator scaled with a scalar."""

    def __init__(self, linop: LinearOperator, scalar: ScalarArgType):
        if not isinstance(linop, LinearOperator):
            raise TypeError("`linop` must be a `LinearOperator`")

        if np.ndim(scalar) != 0:
            raise TypeError("`scalar` must be a scalar.")

        dtype = np.result_type(linop.dtype, scalar)

        self._linop = linop
        self._scalar = probnum.utils.as_numpy_scalar(scalar, dtype)

        super().__init__(
            self._linop.shape,
            dtype=dtype,
            matmul=lambda x: self._scalar * (self._linop @ x),
            rmatmul=lambda x: self._scalar * (x @ self._linop),
            todense=lambda: self._scalar * self._linop.todense(cache=False),
            transpose=lambda: self._scalar * self._linop.T,
            adjoint=lambda: np.conj(self._scalar) * self._linop.H,
            inverse=self._inv,
            trace=lambda: self._scalar * self._linop.trace(),
        )

    def _inv(self) -> "ScaledLinearOperator":
        if self._scalar == 0:
            raise np.linalg.LinAlgError("The operator is not invertible")

        return ScaledLinearOperator(self._linop.inv(), 1.0 / self._scalar)


class NegatedLinearOperator(ScaledLinearOperator):
    def __init__(self, linop: LinearOperator):
        super().__init__(linop, scalar=probnum.utils.as_numpy_scalar(-1, linop.dtype))

    def __neg__(self) -> "LinearOperator":
        return self._linop


class SumLinearOperator(LinearOperator):
    """Sum of two linear operators."""

    def __init__(self, *summands: LinearOperator):
        if not all(isinstance(summand, LinearOperator) for summand in summands):
            raise TypeError("All summands must be `LinearOperator`s")

        if len(summands) < 2:
            raise ValueError("There must be at least two summands")

        if not all(summand.shape == summands[0].shape for summand in summands):
            raise ValueError("All summands must have the same shape")

        self._summands = SumLinearOperator._expand_sum_ops(*summands)

        super().__init__(
            shape=summands[0].shape,
            dtype=np.find_common_type(
                [summand.dtype for summand in self._summands], []
            ),
            matmul=lambda x: functools.reduce(
                operator.add, (summand @ x for summand in self._summands)
            ),
            rmatmul=lambda x: functools.reduce(
                operator.add, (x @ summand for summand in self._summands)
            ),
            todense=lambda: functools.reduce(
                operator.add,
                (summand.todense(cache=False) for summand in self._summands),
            ),
            transpose=lambda: SumLinearOperator(
                *(summand.T for summand in self._summands)
            ),
            adjoint=lambda: SumLinearOperator(
                *(summand.H for summand in self._summands)
            ),
            trace=lambda: functools.reduce(
                operator.add, (summand.trace() for summand in self._summands)
            ),
        )

    def __neg__(self):
        return SumLinearOperator(*(-summand for summand in self._summands))

    @staticmethod
    def _expand_sum_ops(*summands: LinearOperator) -> Tuple[LinearOperator, ...]:
        expanded_summands = []

        for summand in summands:
            if isinstance(summand, SumLinearOperator):
                expanded_summands.extend(summand._summands)
            else:
                expanded_summands.append(summand)

        return tuple(expanded_summands)


def _mul_fallback(
    op1: BinaryOperandType, op2: BinaryOperandType
) -> Union[LinearOperator, type(NotImplemented)]:
    res = NotImplemented

    if isinstance(op1, LinearOperator) and isinstance(op2, LinearOperator):
        pass  # TODO: Implement generic Hadamard product
    elif isinstance(op1, LinearOperator):
        if np.ndim(op2) == 0:
            res = ScaledLinearOperator(op1, op2)
    elif isinstance(op2, LinearOperator):
        if np.ndim(op1) == 0:
            res = ScaledLinearOperator(op2, op1)
    else:
        raise TypeError("At least one of the two operands must be a `LinearOperator`.")

    return res


class ProductLinearOperator(LinearOperator):
    """(Operator) Product of two linear operators."""

    def __init__(self, *factors: LinearOperator):
        if not all(isinstance(factor, LinearOperator) for factor in factors):
            raise TypeError("All factors must be `LinearOperator`s")

        if len(factors) < 2:
            raise ValueError("There must be at least two factors")

        if not all(
            lfactor.shape[1] == rfactor.shape[0]
            for lfactor, rfactor in zip(factors[:-1], factors[1:])
        ):
            raise ValueError(
                f"Shape mismatch: Cannot multiply linear operators with shapes "
                f"{' x '.join(factor.shape for factor in factors)}."
            )

        self._factors = ProductLinearOperator._expand_prod_ops(*factors)

        super().__init__(
            shape=(self._factors[0].shape[0], self._factors[-1].shape[1]),
            dtype=np.find_common_type([factor.dtype for factor in self._factors], []),
            matmul=lambda x: functools.reduce(
                lambda vec, op: op @ vec, reversed(self._factors), x
            ),
            rmatmul=lambda x: functools.reduce(
                lambda vec, op: vec @ op, self._factors, x
            ),
            todense=lambda: functools.reduce(
                operator.matmul,
                (factor.todense(cache=False) for factor in self._factors),
            ),
            transpose=lambda: ProductLinearOperator(
                *(factor.T for factor in reversed(self._factors))
            ),
            adjoint=lambda: ProductLinearOperator(
                *(factor.H for factor in reversed(self._factors))
            ),
            inverse=lambda: ProductLinearOperator(
                *(factor.inv() for factor in reversed(self._factors))
            ),
            det=lambda: functools.reduce(
                operator.mul, (factor.det() for factor in self._factors)
            ),
            logabsdet=lambda: functools.reduce(
                operator.add, (factor.logabsdet() for factor in self._factors)
            ),
        )

    @staticmethod
    def _expand_prod_ops(*factors: LinearOperator) -> Tuple[LinearOperator, ...]:
        expanded_factors = []

        for factor in factors:
            if isinstance(factor, ProductLinearOperator):
                expanded_factors.extend(factor._factors)
            else:
                expanded_factors.append(factor)

        return tuple(expanded_factors)


def _matmul_fallback(
    op1: BinaryOperandType, op2: BinaryOperandType
) -> Union[LinearOperator, type(NotImplemented)]:
    res = NotImplemented

    if isinstance(op1, LinearOperator) and isinstance(op2, LinearOperator):
        res = ProductLinearOperator(op1, op2)

    return res



"""Fixtures to be shared across all modules in this directory.

Mostly some random variables of matching dimensions.
"""

import numpy as np
import pytest

from probnum import randvars
from probnum.problems.zoo.linalg import random_spd_matrix


@pytest.fixture(params=[2])
def test_ndim(request):
    """Test dimension."""
    return request.param


# A few covariance matrices


@pytest.fixture
def spdmat1(test_ndim):
    return random_spd_matrix(test_ndim)


@pytest.fixture
def spdmat2(test_ndim):
    return random_spd_matrix(test_ndim)


@pytest.fixture
def spdmat3(test_ndim):
    return random_spd_matrix(test_ndim)


@pytest.fixture
def spdmat4(test_ndim):
    return random_spd_matrix(test_ndim)


# A few 'Normal' random variables


@pytest.fixture
def some_normal_rv1(test_ndim, spdmat1):

    return randvars.Normal(
        mean=np.random.rand(test_ndim),
        cov=spdmat1,
        cov_cholesky=np.linalg.cholesky(spdmat1),
    )


@pytest.fixture
def some_normal_rv2(test_ndim, spdmat2):
    return randvars.Normal(
        mean=np.random.rand(test_ndim),
        cov=spdmat2,
        cov_cholesky=np.linalg.cholesky(spdmat2),
    )


@pytest.fixture
def some_normal_rv3(test_ndim, spdmat3):
    return randvars.Normal(
        mean=np.random.rand(test_ndim),
        cov=spdmat3,
        cov_cholesky=np.linalg.cholesky(spdmat3),
    )


@pytest.fixture
def some_normal_rv4(test_ndim, spdmat4):
    return randvars.Normal(
        mean=np.random.rand(test_ndim),
        cov=spdmat4,
        cov_cholesky=np.linalg.cholesky(spdmat4),
    )


@pytest.fixture
def diffusion():
    """A diffusion != 1 makes it easier to see if _diffusion is actually used in forward
    and backward."""
    return 5.1412512431


@pytest.fixture(params=["classic", "sqrt"])
def forw_impl_string_linear_gauss(request):
    """Forward implementation choices passed via strings."""
    return request.param


@pytest.fixture(params=["classic", "joseph", "sqrt"])
def backw_impl_string_linear_gauss(request):
    """Backward implementation choices passed via strings."""
    return request.param



"""Test Kalman utility functions."""


import numpy as np
import pytest

import probnum.filtsmooth as pnfs


@pytest.fixture
def d1():
    return 5


@pytest.fixture
def d2():
    return 4


@pytest.fixture
def maxit():
    return 10


@pytest.fixture
def stopcrit():
    return pnfs.StoppingCriterion(atol=1e-1, rtol=1e-1, maxit=10)


def test_continue(stopcrit, d1, d2):
    """Iteration must not terminate if error is large and maxit is not reached."""
    y1 = np.random.rand(d1, d2)
    y2 = 2 * y1.copy() + 3
    current_iterations = stopcrit.iterations
    assert stopcrit.terminate(y1 - y2, y2) is False
    assert stopcrit.iterations == current_iterations + 1


def test_terminate_tolerance(stopcrit, d1, d2):
    """Iteration must terminate if error is small."""
    y1 = np.random.rand(d1, d2)
    y2 = y1.copy() + 1e-8
    assert stopcrit.terminate(y1 - y2, y2) is True
    assert stopcrit.iterations == 0


def test_terminate_maxit(stopcrit, d1, d2, maxit):
    """Iteration must throw an exception if maximum number of iterations is reached."""
    # error is large, which does not really matter...
    y1 = np.random.rand(d1, d2)
    y2 = 2 * y1.copy() + 3
    stopcrit.iterations = maxit + 1
    with pytest.raises(RuntimeError):
        stopcrit.terminate(y1 - y2, y2)



"""Convenience function(s) for state space models."""

import numpy as np
import scipy.stats


def generate_samples(dynmod, measmod, initrv, times, random_state=None):
    """Samples true states and observations at pre-determined timesteps "times" for a
    state space model.

    Parameters
    ----------
    dynmod : statespace.Transition
        Transition model describing the prior dynamics.
    measmod : statespace.Transition
        Transition model describing the measurement model.
    initrv : probnum.RandomVariable object
        Random variable according to initial distribution
    times : np.ndarray, shape (n,)
        Timesteps on which the states are to be sampled.
    random_state :
        Random state that is used to generate the samples from the latent state.
        The measurement samples are not affected by this.

    Returns
    -------
    states : np.ndarray; shape (len(times), dynmod.dimension)
        True states according to dynamic model.
    obs : np.ndarray; shape (len(times), measmod.dimension)
        Observations according to measurement model.
    """
    obs = np.zeros((len(times), measmod.output_dim))

    base_measure_realizations_latent_state = scipy.stats.norm.rvs(
        size=(times.shape + (measmod.input_dim,)), random_state=random_state
    )
    latent_states = np.array(
        dynmod.jointly_transform_base_measure_realization_list_forward(
            base_measure_realizations=base_measure_realizations_latent_state,
            t=times,
            initrv=initrv,
        )
    )

    for idx, (state, t) in enumerate(zip(latent_states, times)):
        measured_rv, _ = measmod.forward_realization(state, t=t)
        obs[idx] = measured_rv.sample()
    return latent_states, obs



"""Initialisation procedures for ODE filters."""
# pylint: disable=import-outside-toplevel


import numpy as np
import scipy.integrate as sci

from probnum import filtsmooth, problems, randvars, statespace

# In the initialisation-via-RK function below, this value is added to the marginal stds of the initial derivatives that are known.
# If we put in zero, there are linalg errors (because a zero-cov RV is conditioned on a dirac likelihood).
# This value is chosen such that its square-root is a really small damping factor).
SMALL_VALUE = 1e-28


def initialize_odefilter_with_rk(
    f, y0, t0, prior, initrv, df=None, h0=1e-2, method="DOP853"
):
    r"""Initialize an ODE filter by fitting the prior process to a few steps of an approximate ODE solution computed with Scipy's RK.

    It goes as follows:

    1. The ODE integration problem is set up on the interval ``[t0, t0 + (2*order+1)*h0]``
    and solved with a call to ``scipy.integrate.solve_ivp``. The solver is uses adaptive steps with ``atol=rtol=1e-12``,
    but is forced to pass through the
    events ``(t0, t0+h0, t0 + 2*h0, ..., t0 + (2*order+1)*h0)``.
    The result is a vector of time points and states, with at least ``(2*order+1)``.
    Potentially, the adaptive steps selected many more steps, but because of the events, fewer steps cannot have happened.

    2. A prescribed prior is fitted to the first ``(2*order+1)`` (t, y) pairs of the solution. ``order`` is the order of the prior.

    3. The value of the resulting posterior at time ``t=t0`` is an estimate of the state and all its derivatives.
    The resulting marginal standard deviations estimate the error. This random variable is returned.

    Parameters
    ----------
    f
        ODE vector field.
    y0
        Initial value.
    t0
        Initial time point.
    prior
        Prior distribution used for the ODE solver. For instance an integrated Brownian motion prior (``IBM``).
    initrv
        Initial random variable.
    df
        Jacobian of the ODE vector field. Optional. If specified, more components of the result will be exact.
    h0
        Maximum step-size to use for computing the approximate ODE solution. The smaller, the more accurate, but also, the smaller, the less stable.
        The best value here depends on the ODE problem, and probably the chosen method. Optional. Default is ``1e-2``.
    method
        Which solver to use. This is communicated as a string that is compatible with ``scipy.integrate.solve_ivp(..., method=method)``.
        Optional. Default is `DOP853`.

    Returns
    -------
    Normal
        Estimated (improved) initial random variable. Compatible with the specified prior.

    Examples
    --------

    >>> from dataclasses import astuple
    >>> from probnum.randvars import Normal
    >>> from probnum.statespace import IBM
    >>> from probnum.problems.zoo.diffeq import vanderpol

    Compute the initial values of the van-der-Pol problem as follows

    >>> f, t0, tmax, y0, df, *_ = astuple(vanderpol())
    >>> print(y0)
    [2. 0.]
    >>> prior = IBM(ordint=3, spatialdim=2)
    >>> initrv = Normal(mean=np.zeros(prior.dimension), cov=np.eye(prior.dimension))
    >>> improved_initrv = initialize_odefilter_with_rk(f, y0, t0, prior=prior, initrv=initrv, df=df)
    >>> print(prior.proj2coord(0) @ improved_initrv.mean)
    [2. 0.]
    >>> print(np.round(improved_initrv.mean, 1))
    [    2.      0.     -2.     58.2     0.     -2.     60.  -1745.7]
    >>> print(np.round(np.log10(improved_initrv.std), 1))
    [-13.8 -11.3  -9.   -1.5 -13.8 -11.3  -9.   -1.5]
    """
    y0 = np.asarray(y0)
    ode_dim = y0.shape[0] if y0.ndim > 0 else 1
    order = prior.ordint
    proj_to_y = prior.proj2coord(0)
    zeros_shift = np.zeros(ode_dim)
    zeros_cov = np.zeros((ode_dim, ode_dim))
    measmod = statespace.DiscreteLTIGaussian(
        proj_to_y,
        zeros_shift,
        zeros_cov,
        proc_noise_cov_cholesky=zeros_cov,
        forward_implementation="sqrt",
        backward_implementation="sqrt",
    )

    # order + 1 would suffice in theory, 2*order + 1 is for good measure
    # (the "+1" is a safety factor for order=1)
    num_steps = 2 * order + 1
    t_eval = np.arange(t0, t0 + (num_steps + 1) * h0, h0)
    sol = sci.solve_ivp(
        f,
        (t0, t0 + (num_steps + 1) * h0),
        y0=y0,
        atol=1e-12,
        rtol=1e-12,
        t_eval=t_eval,
        method=method,
    )

    ts = sol.t[:num_steps]
    ys = sol.y[:, :num_steps].T

    initmean = initrv.mean.copy()
    initmean[0 :: (order + 1)] = y0
    initmean[1 :: (order + 1)] = f(t0, y0)

    initcov_diag = np.diag(initrv.cov).copy()
    initcov_diag[0 :: (order + 1)] = SMALL_VALUE
    initcov_diag[1 :: (order + 1)] = SMALL_VALUE

    if df is not None:
        if order > 1:
            initmean[2 :: (order + 1)] = df(t0, y0) @ f(t0, y0)
            initcov_diag[2 :: (order + 1)] = SMALL_VALUE

    initcov = np.diag(initcov_diag)
    initcov_cholesky = np.diag(np.sqrt(initcov_diag))
    initrv = randvars.Normal(initmean, initcov, cov_cholesky=initcov_cholesky)
    kalman = filtsmooth.Kalman(prior, measmod, initrv)

    regression_problem = problems.RegressionProblem(observations=ys, locations=ts)
    out, _ = kalman.filtsmooth(regression_problem)

    estimated_initrv = out.states[0]

    return estimated_initrv


def initialize_odefilter_with_taylormode(f, y0, t0, prior, initrv):
    """Initialize an ODE filter with Taylor-mode automatic differentiation.

    This requires JAX. For an explanation of what happens ``under the hood``, see [1]_.

    References
    ----------
    .. [1] Krämer, N. and Hennig, P., Stable implementation of probabilistic ODE solvers,
       *arXiv:2012.10106*, 2020.


    The implementation is inspired by the implementation in
    https://github.com/jacobjinkelly/easy-neural-ode/blob/master/latent_ode.py

    Parameters
    ----------
    f
        ODE vector field.
    y0
        Initial value.
    t0
        Initial time point.
    prior
        Prior distribution used for the ODE solver. For instance an integrated Brownian motion prior (``IBM``).
    initrv
        Initial random variable.

    Returns
    -------
    Normal
        Estimated initial random variable. Compatible with the specified prior.


    Examples
    --------

    >>> import sys, pytest
    >>> if not sys.platform.startswith('linux'):
    ...     pytest.skip()

    >>> from dataclasses import astuple
    >>> from probnum.randvars import Normal
    >>> from probnum.problems.zoo.diffeq import threebody_jax, vanderpol_jax
    >>> from probnum.statespace import IBM

    Compute the initial values of the restricted three-body problem as follows

    >>> f, t0, tmax, y0, df, *_ = astuple(threebody_jax())
    >>> print(y0)
    [ 0.994       0.          0.         -2.00158511]

    >>> prior = IBM(ordint=3, spatialdim=4)
    >>> initrv = Normal(mean=np.zeros(prior.dimension), cov=np.eye(prior.dimension))
    >>> improved_initrv = initialize_odefilter_with_taylormode(f, y0, t0, prior, initrv)
    >>> print(prior.proj2coord(0) @ improved_initrv.mean)
    [ 0.994       0.          0.         -2.00158511]
    >>> print(improved_initrv.mean)
    [ 9.94000000e-01  0.00000000e+00 -3.15543023e+02  0.00000000e+00
      0.00000000e+00 -2.00158511e+00  0.00000000e+00  9.99720945e+04
      0.00000000e+00 -3.15543023e+02  0.00000000e+00  6.39028111e+07
     -2.00158511e+00  0.00000000e+00  9.99720945e+04  0.00000000e+00]

    Compute the initial values of the van-der-Pol oscillator as follows

    >>> f, t0, tmax, y0, df, *_ = astuple(vanderpol_jax())
    >>> print(y0)
    [2. 0.]
    >>> prior = IBM(ordint=3, spatialdim=2)
    >>> initrv = Normal(mean=np.zeros(prior.dimension), cov=np.eye(prior.dimension))
    >>> improved_initrv = initialize_odefilter_with_taylormode(f, y0, t0, prior, initrv)
    >>> print(prior.proj2coord(0) @ improved_initrv.mean)
    [2. 0.]
    >>> print(improved_initrv.mean)
    [    2.     0.    -2.    60.     0.    -2.    60. -1798.]
    >>> print(improved_initrv.std)
    [0. 0. 0. 0. 0. 0. 0. 0.]
    """

    try:
        import jax.numpy as jnp
        from jax.config import config
        from jax.experimental.jet import jet

        config.update("jax_enable_x64", True)
    except ImportError as err:
        raise ImportError(
            "Cannot perform Taylor-mode initialisation without optional "
            "dependencies jax and jaxlib. Try installing them via `pip install jax jaxlib`."
        ) from err

    order = prior.ordint

    def total_derivative(z_t):
        """Total derivative."""
        z, t = jnp.reshape(z_t[:-1], z_shape), z_t[-1]
        dz = jnp.ravel(f(t, z))
        dt = jnp.array([1.0])
        dz_t = jnp.concatenate((dz, dt))
        return dz_t

    z_shape = y0.shape
    z_t = jnp.concatenate((jnp.ravel(y0), jnp.array([t0])))

    derivs = []

    derivs.extend(y0)
    if order == 0:
        all_derivs = statespace.Integrator._convert_derivwise_to_coordwise(
            np.asarray(jnp.array(derivs)), ordint=0, spatialdim=len(y0)
        )

        return randvars.Normal(
            np.asarray(all_derivs),
            cov=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
            cov_cholesky=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
        )

    (dy0, [*yns]) = jet(total_derivative, (z_t,), ((jnp.ones_like(z_t),),))
    derivs.extend(dy0[:-1])
    if order == 1:
        all_derivs = statespace.Integrator._convert_derivwise_to_coordwise(
            np.asarray(jnp.array(derivs)), ordint=1, spatialdim=len(y0)
        )

        return randvars.Normal(
            np.asarray(all_derivs),
            cov=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
            cov_cholesky=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
        )

    for _ in range(1, order):
        (dy0, [*yns]) = jet(total_derivative, (z_t,), ((dy0, *yns),))
        derivs.extend(yns[-2][:-1])

    all_derivs = statespace.Integrator._convert_derivwise_to_coordwise(
        jnp.array(derivs), ordint=order, spatialdim=len(y0)
    )

    return randvars.Normal(
        np.asarray(all_derivs),
        cov=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
        cov_cholesky=np.asarray(jnp.diag(jnp.zeros(len(derivs)))),
    )



"""Random processes."""

from ._gaussian_process import GaussianProcess
from ._markov_process import MarkovProcess
from ._random_process import RandomProcess

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "RandomProcess",
    "MarkovProcess",
    "GaussianProcess",
]

# Set correct module paths. Corrects links and module paths in documentation.
RandomProcess.__module__ = "probnum.randprocs"
MarkovProcess.__module__ = "probnum.randprocs"
GaussianProcess.__module__ = "probnum.randprocs"



import unittest

import numpy as np

from probnum._randomvariablelist import _RandomVariableList
from probnum.randvars import Constant


class TestRandomVariableList(unittest.TestCase):
    def setUp(self):
        self.rv_list = _RandomVariableList([Constant(0.1), Constant(0.2)])

    def test_inputs(self):
        """Inputs rejected or accepted according to expected types."""
        numpy_array = np.ones(3) * Constant(0.1)
        constant_list = [Constant(0.1), Constant(0.4)]
        number_list = [0.5, 0.41]
        inputs = [numpy_array, constant_list, number_list]
        inputs_acceptable = [False, True, False]

        for inputs, is_acceptable in zip(inputs, inputs_acceptable):
            with self.subTest(input=inputs, is_acceptable=is_acceptable):

                if is_acceptable:
                    _RandomVariableList(inputs)
                else:
                    with self.assertRaises(TypeError):
                        _RandomVariableList(inputs)

    def test_mean(self):
        mean = self.rv_list.mean
        self.assertEqual(mean.shape, (2,))

    def test_cov(self):
        cov = self.rv_list.cov
        self.assertEqual(cov.shape, (2,))

    def test_var(self):
        var = self.rv_list.var
        self.assertEqual(var.shape, (2,))

    def test_std(self):
        std = self.rv_list.std
        self.assertEqual(std.shape, (2,))

    def test_getitem(self):
        item = self.rv_list[0]
        self.assertIsInstance(item, Constant)

    def test_shape(self):
        self.assertEqual(self.rv_list.shape, (2,))


class TestEmptyRandomVariableList(unittest.TestCase):
    """Passing an empty list does not screw things up."""

    def setUp(self):
        self.rv_list = _RandomVariableList([])

    def test_mean(self):
        mean = self.rv_list.mean
        self.assertEqual(mean.shape, (0,))

    def test_cov(self):
        cov = self.rv_list.cov
        self.assertEqual(cov.shape, (0,))

    def test_var(self):
        var = self.rv_list.var
        self.assertEqual(var.shape, (0,))

    def test_std(self):
        std = self.rv_list.std
        self.assertEqual(std.shape, (0,))

    def test_getitem(self):
        with self.assertRaises(IndexError):
            item = self.rv_list[0]


if __name__ == "__main__":
    unittest.main()



"""Kernels or covariance functions."""

from ._exponentiated_quadratic import ExpQuad
from ._kernel import Kernel
from ._linear import Linear
from ._matern import Matern
from ._polynomial import Polynomial
from ._rational_quadratic import RatQuad
from ._white_noise import WhiteNoise

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "Kernel",
    "WhiteNoise",
    "Linear",
    "Polynomial",
    "ExpQuad",
    "RatQuad",
    "Matern",
]

# Set correct module paths. Corrects links and module paths in documentation.
Kernel.__module__ = "probnum.kernels"
WhiteNoise.__module__ = "probnum.kernels"
Linear.__module__ = "probnum.kernels"
Polynomial.__module__ = "probnum.kernels"
ExpQuad.__module__ = "probnum.kernels"
RatQuad.__module__ = "probnum.kernels"
Matern.__module__ = "probnum.kernels"



"""Interface for transition test functions."""

import abc


class InterfaceTestTransition(abc.ABC):
    @abc.abstractmethod
    def test_forward_rv(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def test_forward_realization(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def test_backward_rv(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def test_backward_realization(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def test_input_dim(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def test_output_dim(self, *args, **kwargs):
        raise NotImplementedError



"""Tests for constant random variables."""
import unittest

import numpy as np

from probnum import randvars


def test_constant_accessible_like_gaussian():
    """Constant has an attribute cov_cholesky which returns zeros."""
    support = np.array([2, 3])
    s = randvars.Constant(support)
    np.testing.assert_allclose(s.cov, s.cov_cholesky)


class TestConstant(unittest.TestCase):
    """General test case for constant random variables."""

    def setUp(self):
        self.supports = [1, np.array([1, 2]), np.array([[0]]), np.array([[6], [-0.3]])]

    def test_logpdf(self):
        pass

    def test_sample_shapes(self):
        """Test whether samples have the correct shapes."""
        for supp in self.supports:
            for sample_size in [1, (), 10, (4,), (3, 2)]:
                with self.subTest():
                    s = randvars.Constant(support=supp).sample(size=sample_size)
                    if sample_size == ():
                        self.assertEqual(np.shape(supp), np.shape(s))
                    elif isinstance(sample_size, tuple):
                        self.assertEqual(sample_size + np.shape(supp), np.shape(s))
                    else:
                        self.assertEqual(
                            tuple([sample_size, *np.shape(supp)]), np.shape(s)
                        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pytest

from probnum import filtsmooth, randvars


@pytest.fixture
def state_list():
    return [
        randvars.Categorical(
            probabilities=np.ones(10) / 10, support=np.random.rand(10, 2)
        )
        for _ in range(20)
    ]


@pytest.fixture
def posterior(state_list):
    return filtsmooth.ParticleFilterPosterior(
        states=state_list, locations=np.random.rand(20)
    )


def test_len(posterior):
    assert len(posterior) == 20


def test_getitem(posterior):
    assert len(posterior) == 20


def test_call(posterior):
    with pytest.raises(NotImplementedError):
        posterior(0.0)


# The tests below -- as a side results -- also cover the specific properties in _RandomVariableList.


def test_mode(posterior):
    mode = posterior.states.mode
    assert mode.shape == (len(posterior),) + posterior.states[0].shape


def test_support(posterior):
    support = posterior.states.support
    assert (
        support.shape
        == (len(posterior),)
        + (len(posterior.states[0].probabilities),)
        + posterior.states[0].shape
    )


def test_probabilities(posterior):
    probabilities = posterior.states.probabilities
    assert probabilities.shape == (len(posterior),) + (
        len(posterior.states[0].probabilities),
    )


def test_resample(posterior):
    resampled_posterior = posterior.states.resample()
    assert (
        resampled_posterior.support.shape
        == (len(posterior),)
        + (len(posterior.states[0].probabilities),)
        + posterior.states[0].shape
    )



"""Normally distributed / Gaussian random variables."""

from typing import Callable, Optional, Union

import numpy as np
import scipy.linalg
import scipy.stats

from probnum import linops
from probnum import utils as _utils
from probnum.type import (
    ArrayLikeGetitemArgType,
    FloatArgType,
    RandomStateArgType,
    ShapeArgType,
    ShapeType,
)

from . import _random_variable

try:
    # functools.cached_property is only available in Python >=3.8
    from functools import cached_property
except ImportError:
    from cached_property import cached_property


COV_CHOLESKY_DAMPING = 10 ** -12

_ValueType = Union[np.floating, np.ndarray, linops.LinearOperator]


class Normal(_random_variable.ContinuousRandomVariable[_ValueType]):
    """Random variable with a normal distribution.

    Gaussian random variables are ubiquitous in probability theory, since the
    Gaussian is the equilibrium distribution to which other distributions gravitate
    under a wide variety of smooth operations, e.g., convolutions and stochastic
    transformations. One example of this is the central limit theorem. Gaussian
    random variables are also attractive from a numerical point of view as they
    maintain their distribution family through many transformations (e.g. they are
    stable). In particular, they allow for efficient closed-form Bayesian inference
    given linear observations.

    Parameters
    ----------
    mean :
        Mean of the random variable.
    cov :
        (Co-)variance of the random variable.
    cov_cholesky :
        (Lower triangular) Cholesky factor of the covariance matrix. If None, then the Cholesky factor of the covariance matrix
        is computed when :attr:`Normal.cov_cholesky` is called and then cached. If specified, the value is returned by :attr:`Normal.cov_cholesky`.
        In this case, its type and data type are compared to the type and data type of the covariance.
        If the types do not match, an exception is thrown. If the data types do not match,
        the data type of the Cholesky factor is promoted to the data type of the covariance matrix.
    random_state :
        Random state of the random variable. If None (or np.random), the global
        :mod:`numpy.random` state is used. If integer, it is used to seed the local
        :class:`~numpy.random.RandomState` instance.

    See Also
    --------
    RandomVariable : Class representing random variables.

    Examples
    --------
    >>> from probnum import randvars
    >>> x = randvars.Normal(mean=0.5, cov=1.0, random_state=42)
    >>> x.sample(size=(2, 2))
    array([[0.99671415, 0.3617357 ],
           [1.14768854, 2.02302986]])
    """

    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def __init__(
        self,
        mean: Union[float, np.floating, np.ndarray, linops.LinearOperator],
        cov: Union[float, np.floating, np.ndarray, linops.LinearOperator],
        cov_cholesky: Optional[
            Union[float, np.floating, np.ndarray, linops.LinearOperator]
        ] = None,
        random_state: RandomStateArgType = None,
    ):
        # Type normalization
        if np.isscalar(mean):
            mean = _utils.as_numpy_scalar(mean)

        if np.isscalar(cov):
            cov = _utils.as_numpy_scalar(cov)

        if np.isscalar(cov_cholesky):
            cov_cholesky = _utils.as_numpy_scalar(cov_cholesky)

        # Data type normalization
        dtype = np.promote_types(mean.dtype, cov.dtype)

        if not np.issubdtype(dtype, np.floating):
            dtype = np.dtype(np.double)

        mean = mean.astype(dtype, order="C", casting="safe", copy=False)
        cov = cov.astype(dtype, order="C", casting="safe", copy=False)

        # Shape checking
        if not 0 <= mean.ndim <= 2:
            raise ValueError(
                f"Gaussian random variables must either be scalars, vectors, or "
                f"matrices (or linear operators), but the given mean is a {mean.ndim}-"
                f"dimensional tensor."
            )

        expected_cov_shape = (np.prod(mean.shape),) * 2 if len(mean.shape) > 0 else ()

        if cov.shape != expected_cov_shape:
            raise ValueError(
                f"The covariance matrix must be of shape {expected_cov_shape}, but "
                f"shape {cov.shape} was given."
            )

        self._mean = mean
        self._cov = cov

        self._compute_cov_cholesky: Callable[[], _ValueType] = None
        self._cov_cholesky = cov_cholesky  # recall: None if not provided

        # Method selection
        univariate = mean.ndim == 0
        dense = isinstance(mean, np.ndarray) and isinstance(cov, np.ndarray)
        cov_operator = isinstance(cov, linops.LinearOperator)

        if univariate:
            # Univariate Gaussian
            sample = self._univariate_sample
            in_support = Normal._univariate_in_support
            pdf = self._univariate_pdf
            logpdf = self._univariate_logpdf
            cdf = self._univariate_cdf
            logcdf = self._univariate_logcdf
            quantile = self._univariate_quantile

            median = lambda: self._mean
            var = lambda: self._cov
            entropy = self._univariate_entropy

            self._compute_cov_cholesky = self._univariate_cov_cholesky

        elif dense or cov_operator:
            # Multi- and matrixvariate Gaussians
            sample = self._dense_sample
            in_support = Normal._dense_in_support
            pdf = self._dense_pdf
            logpdf = self._dense_logpdf
            cdf = self._dense_cdf
            logcdf = self._dense_logcdf
            quantile = None

            median = None
            var = self._dense_var
            entropy = self._dense_entropy

            self._compute_cov_cholesky = self.dense_cov_cholesky

            # Ensure that the Cholesky factor has the same type as the covariance,
            # and, if necessary, promote data types. Check for (in this order): type, shape, dtype.
            if self._cov_cholesky is not None:

                if not isinstance(self._cov_cholesky, type(self._cov)):
                    raise TypeError(
                        f"The covariance matrix is of type `{type(self._cov)}`, so its "
                        f"Cholesky decomposition must be of the same type, but an "
                        f"object of type `{type(self._cov_cholesky)}` was given."
                    )

                if self._cov_cholesky.shape != self._cov.shape:
                    raise ValueError(
                        f"The cholesky decomposition of the covariance matrix must "
                        f"have the same shape as the covariance matrix, i.e. "
                        f"{self._cov.shape}, but shape {self._cov_cholesky.shape} was given"
                    )

                if self._cov_cholesky.dtype != self._cov.dtype:
                    self._cov_cholesky = self._cov_cholesky.astype(
                        self._cov.dtype, casting="safe", copy=False
                    )

            if isinstance(cov, linops.SymmetricKronecker):
                m, n = mean.shape

                if m != n or n != cov.A.shape[0] or n != cov.B.shape[1]:
                    raise ValueError(
                        "Normal distributions with symmetric Kronecker structured "
                        "kernels must have square mean and square kernels factors with "
                        "matching dimensions."
                    )

                if cov.identical_factors:
                    sample = self._symmetric_kronecker_identical_factors_sample

                    # pylint: disable=redefined-variable-type
                    self._compute_cov_cholesky = (
                        self._symmetric_kronecker_identical_factors_cov_cholesky
                    )
            elif isinstance(cov, linops.Kronecker):
                m, n = mean.shape

                if (
                    m != cov.A.shape[0]
                    or m != cov.A.shape[1]
                    or n != cov.B.shape[0]
                    or n != cov.B.shape[1]
                ):
                    raise ValueError(
                        "Kronecker structured kernels must have factors with the same "
                        "shape as the mean."
                    )

                self._compute_cov_cholesky = self._kronecker_cov_cholesky

        else:
            raise ValueError(
                f"Cannot instantiate normal distribution with mean of type "
                f"{mean.__class__.__name__} and kernels of type "
                f"{cov.__class__.__name__}."
            )

        super().__init__(
            shape=mean.shape,
            dtype=mean.dtype,
            random_state=random_state,
            parameters={"mean": self._mean, "cov": self._cov},
            sample=sample,
            in_support=in_support,
            pdf=pdf,
            logpdf=logpdf,
            cdf=cdf,
            logcdf=logcdf,
            quantile=quantile,
            mode=lambda: self._mean,
            median=median,
            mean=lambda: self._mean,
            cov=lambda: self._cov,
            var=var,
            entropy=entropy,
        )

    @property
    def cov_cholesky(self) -> _ValueType:
        """Cholesky factor :math:`L` of the covariance
        :math:`\\operatorname{Cov}(X) =LL^\\top`."""

        if not self.cov_cholesky_is_precomputed:
            self.precompute_cov_cholesky()
        return self._cov_cholesky

    def precompute_cov_cholesky(
        self, damping_factor: Optional[FloatArgType] = COV_CHOLESKY_DAMPING
    ):
        """(P)recompute Cholesky factors (careful: in-place operation!)."""
        if self.cov_cholesky_is_precomputed:
            raise Exception("A Cholesky factor is already available.")
        self._cov_cholesky = self._compute_cov_cholesky(damping_factor=damping_factor)

    @property
    def cov_cholesky_is_precomputed(self) -> bool:
        """Return truth-value of whether the Cholesky factor of the covariance is
        readily available.

        This happens if (i) the Cholesky factor is specified during
        initialization or if (ii) the property `self.cov_cholesky` has
        been called before.
        """
        if self._cov_cholesky is None:
            return False
        return True

    @cached_property
    def dense_mean(self) -> Union[np.floating, np.ndarray]:
        """Dense representation of the mean."""
        if isinstance(self._mean, linops.LinearOperator):
            return self._mean.todense()
        else:
            return self._mean

    @cached_property
    def dense_cov(self) -> Union[np.floating, np.ndarray]:
        """Dense representation of the covariance."""
        if isinstance(self._cov, linops.LinearOperator):
            return self._cov.todense()
        else:
            return self._cov

    def __getitem__(self, key: ArrayLikeGetitemArgType) -> "Normal":
        """Marginalization in multi- and matrixvariate normal random variables,
        expressed as (advanced) indexing, masking and slicing.

        We support all modes of array indexing presented in

        https://numpy.org/doc/1.19/reference/arrays.indexing.html.

        Note that, currently, this method only works for multi- and matrixvariate
        normal distributions.

        Parameters
        ----------
        key : int or slice or ndarray or tuple of None, int, slice, or ndarray
            Indices, slice objects and/or boolean masks specifying which entries to keep
            while marginalizing over all other entries.
        """

        if not isinstance(key, tuple):
            key = (key,)

        # Select entries from mean
        mean = self.dense_mean[key]

        # Select submatrix from covariance matrix
        cov = self.dense_cov.reshape(self.shape + self.shape)
        cov = cov[key][(...,) + key]

        if mean.ndim > 0:
            cov = cov.reshape(mean.size, mean.size)

        return Normal(
            mean=mean,
            cov=cov,
            random_state=_utils.derive_random_seed(self.random_state),
        )

    def reshape(self, newshape: ShapeArgType) -> "Normal":
        try:
            reshaped_mean = self.dense_mean.reshape(newshape)
        except ValueError as exc:
            raise ValueError(
                f"Cannot reshape this normal random variable to the given shape: "
                f"{newshape}"
            ) from exc

        reshaped_cov = self.dense_cov

        if reshaped_mean.ndim > 0 and reshaped_cov.ndim == 0:
            reshaped_cov = reshaped_cov.reshape(1, 1)

        return Normal(
            mean=reshaped_mean,
            cov=reshaped_cov,
            random_state=_utils.derive_random_seed(self.random_state),
        )

    def transpose(self, *axes: int) -> "Normal":
        if len(axes) == 1 and isinstance(axes[0], tuple):
            axes = axes[0]
        elif (len(axes) == 1 and axes[0] is None) or len(axes) == 0:
            axes = tuple(reversed(range(self.ndim)))

        mean_t = self.dense_mean.transpose(*axes).copy()

        # Transpose covariance
        cov_axes = axes + tuple(mean_t.ndim + axis for axis in axes)
        cov_t = self.dense_cov.reshape(self.shape + self.shape)
        cov_t = cov_t.transpose(*cov_axes).copy()

        if mean_t.ndim > 0:
            cov_t = cov_t.reshape(mean_t.size, mean_t.size)

        return Normal(
            mean=mean_t,
            cov=cov_t,
            random_state=_utils.derive_random_seed(self.random_state),
        )

    # Unary arithmetic operations

    def __neg__(self) -> "Normal":
        return Normal(
            mean=-self._mean,
            cov=self._cov,
            random_state=_utils.derive_random_seed(self.random_state),
        )

    def __pos__(self) -> "Normal":
        return Normal(
            mean=+self._mean,
            cov=self._cov,
            random_state=_utils.derive_random_seed(self.random_state),
        )

    # TODO: Overwrite __abs__ and add absolute moments of normal
    # TODO: (https://arxiv.org/pdf/1209.4340.pdf)

    # Binary arithmetic operations

    def _add_normal(self, other: "Normal") -> "Normal":
        if other.shape != self.shape:
            raise ValueError(
                "Addition of two normally distributed random variables is only "
                "possible if both operands have the same shape."
            )

        return Normal(
            mean=self._mean + other._mean,
            cov=self._cov + other._cov,
            random_state=_utils.derive_random_seed(
                self.random_state, other.random_state
            ),
        )

    def _sub_normal(self, other: "Normal") -> "Normal":
        if other.shape != self.shape:
            raise ValueError(
                "Subtraction of two normally distributed random variables is only "
                "possible if both operands have the same shape."
            )

        return Normal(
            mean=self._mean - other._mean,
            cov=self._cov + other._cov,
            random_state=_utils.derive_random_seed(
                self.random_state, other.random_state
            ),
        )

    # Univariate Gaussians
    def _univariate_cov_cholesky(
        self, damping_factor: Optional[FloatArgType] = COV_CHOLESKY_DAMPING
    ) -> np.floating:
        return np.sqrt(self._cov + damping_factor)

    def _univariate_sample(
        self, size: ShapeType = ()
    ) -> Union[np.floating, np.ndarray]:
        sample = scipy.stats.norm.rvs(
            loc=self._mean, scale=self.std, size=size, random_state=self.random_state
        )

        if np.isscalar(sample):
            sample = _utils.as_numpy_scalar(sample, dtype=self.dtype)
        else:
            sample = sample.astype(self.dtype)

        assert sample.shape == size

        return sample

    @staticmethod
    def _univariate_in_support(x: _ValueType) -> bool:
        return np.isfinite(x)

    def _univariate_pdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.norm.pdf(x, loc=self._mean, scale=self.std)

    def _univariate_logpdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.norm.logpdf(x, loc=self._mean, scale=self.std)

    def _univariate_cdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.norm.cdf(x, loc=self._mean, scale=self.std)

    def _univariate_logcdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.norm.logcdf(x, loc=self._mean, scale=self.std)

    def _univariate_quantile(self, p: FloatArgType) -> np.floating:
        return scipy.stats.norm.ppf(p, loc=self._mean, scale=self.std)

    def _univariate_entropy(self: _ValueType) -> np.float_:
        return _utils.as_numpy_scalar(
            scipy.stats.norm.entropy(loc=self._mean, scale=self.std),
            dtype=np.float_,
        )

    # Multi- and matrixvariate Gaussians
    def dense_cov_cholesky(
        self, damping_factor: Optional[FloatArgType] = COV_CHOLESKY_DAMPING
    ) -> np.ndarray:
        """Compute the Cholesky factorization of the covariance from its dense
        representation."""
        dense_cov = self.dense_cov

        return scipy.linalg.cholesky(
            dense_cov + damping_factor * np.eye(self.size, dtype=self.dtype),
            lower=True,
        )

    def _dense_sample(self, size: ShapeType = ()) -> np.ndarray:
        sample = scipy.stats.multivariate_normal.rvs(
            mean=self.dense_mean.ravel(),
            cov=self.dense_cov,
            size=size,
            random_state=self.random_state,
        )

        return sample.reshape(sample.shape[:-1] + self.shape)

    @staticmethod
    def _arg_todense(x: Union[np.ndarray, linops.LinearOperator]) -> np.ndarray:
        if isinstance(x, linops.LinearOperator):
            return x.todense()
        elif isinstance(x, np.ndarray):
            return x
        else:
            raise ValueError(f"Unsupported argument type {type(x)}")

    @staticmethod
    def _dense_in_support(x: _ValueType) -> bool:
        return np.all(np.isfinite(Normal._arg_todense(x)))

    def _dense_pdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.multivariate_normal.pdf(
            Normal._arg_todense(x).reshape(x.shape[: -self.ndim] + (-1,)),
            mean=self.dense_mean.ravel(),
            cov=self.dense_cov,
        )

    def _dense_logpdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.multivariate_normal.logpdf(
            Normal._arg_todense(x).reshape(x.shape[: -self.ndim] + (-1,)),
            mean=self.dense_mean.ravel(),
            cov=self.dense_cov,
        )

    def _dense_cdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.multivariate_normal.cdf(
            Normal._arg_todense(x).reshape(x.shape[: -self.ndim] + (-1,)),
            mean=self.dense_mean.ravel(),
            cov=self.dense_cov,
        )

    def _dense_logcdf(self, x: _ValueType) -> np.float_:
        return scipy.stats.multivariate_normal.logcdf(
            Normal._arg_todense(x).reshape(x.shape[: -self.ndim] + (-1,)),
            mean=self.dense_mean.ravel(),
            cov=self.dense_cov,
        )

    def _dense_var(self) -> np.ndarray:
        return np.diag(self.dense_cov).reshape(self.shape)

    def _dense_entropy(self) -> np.float_:
        return _utils.as_numpy_scalar(
            scipy.stats.multivariate_normal.entropy(
                mean=self.dense_mean.ravel(),
                cov=self.dense_cov,
            ),
            dtype=np.float_,
        )

    # Matrixvariate Gaussian with Kronecker covariance
    def _kronecker_cov_cholesky(
        self, damping_factor: Optional[FloatArgType] = COV_CHOLESKY_DAMPING
    ) -> linops.Kronecker:
        assert isinstance(self._cov, linops.Kronecker)

        A = self._cov.A.todense()
        B = self._cov.B.todense()

        return linops.Kronecker(
            A=scipy.linalg.cholesky(
                A + damping_factor * np.eye(A.shape[0], dtype=self.dtype),
                lower=True,
            ),
            B=scipy.linalg.cholesky(
                B + damping_factor * np.eye(B.shape[0], dtype=self.dtype),
                lower=True,
            ),
        )

    # Matrixvariate Gaussian with symmetric Kronecker covariance from identical
    # factors
    def _symmetric_kronecker_identical_factors_cov_cholesky(
        self,
        damping_factor: Optional[FloatArgType] = COV_CHOLESKY_DAMPING,
    ) -> linops.SymmetricKronecker:
        assert (
            isinstance(self._cov, linops.SymmetricKronecker)
            and self._cov.identical_factors
        )

        A = self._cov.A.todense()

        return linops.SymmetricKronecker(
            A=scipy.linalg.cholesky(
                A + damping_factor * np.eye(A.shape[0], dtype=self.dtype),
                lower=True,
            ),
        )

    def _symmetric_kronecker_identical_factors_sample(
        self, size: ShapeType = ()
    ) -> np.ndarray:
        assert (
            isinstance(self._cov, linops.SymmetricKronecker)
            and self._cov.identical_factors
        )

        n = self._mean.shape[1]

        # Draw standard normal samples
        size_sample = (n * n,) + size

        stdnormal_samples = scipy.stats.norm.rvs(
            size=size_sample, random_state=self.random_state
        )

        # Appendix E: Bartels, S., Probabilistic Linear Algebra, PhD Thesis 2019
        samples_scaled = linops.Symmetrize(n) @ (self.cov_cholesky @ stdnormal_samples)

        # TODO: can we avoid todense here and just return operator samples?
        return self.dense_mean[None, :, :] + samples_scaled.T.reshape(-1, n, n)



"""Sparse matrices from the SuiteSparse Matrix Collection."""

import codecs
import csv
import io
import tarfile
from typing import Dict, Union

import numpy as np
import scipy.io

import probnum.linops as linops

# URLs and file paths to data
SUITESPARSE_ROOT_URL = "https://sparse.tamu.edu"
SUITESPARSE_INDEX_URL = SUITESPARSE_ROOT_URL + "/files/ssstats.csv"


def suitesparse_matrix(
    name: str,
    group: str,
    verbose: bool = False,
) -> "SuiteSparseMatrix":
    """Sparse matrix from the SuiteSparse Matrix Collection.

    Download a sparse matrix benchmark from the `SuiteSparse Matrix Collection
    <https://sparse.tamu.edu/>`_. [1]_ [2]_

    Parameters
    ----------
    name :
        Name of the matrix.
    group :
        Group of the matrix.
    verbose :
        Print additional information.

    References
    ----------
    .. [1] Davis, TA and Hu, Y. The University of Florida sparse matrix
           collection. *ACM Transactions on Mathematical Software (TOMS)* 38.1
           (2011): 1-25.
    .. [2] Kolodziej, Scott P., et al. The SuiteSparse matrix collection website
           interface. *Journal of Open Source Software* 4.35 (2019): 1244.

    Examples
    --------
    >>> ssmat = suitesparse_matrix(name="ash85", group="HB")
    >>> ssmat
    <SuiteSparseMatrix with shape=(85, 85) and dtype=float64>
    >>> ssmat.trace()
    85.0
    """
    # Get database index
    try:
        import requests  # pylint: disable=import-outside-toplevel
    except ImportError as err:
        raise ImportError(
            "Cannot query SuiteSparse Matrix collection without optional dependency ",
            "`requests`. Install ProbNum with optional dependencies for the problem zoo via",
            " `pip install probnum[zoo]` or install requests directly: `pip install requests`.",
        ) from err
    response = requests.get(SUITESPARSE_INDEX_URL, "r")
    line_gen = response.iter_lines()
    for _ in range(2):
        next(line_gen)  # skip lines not part of the matrix table

    # Read index with custom header
    fieldnames = [
        "group",
        "name",
        "nrows",
        "ncols",
        "nnz",
        "real",
        "logical",
        "is2d3d",
        "isspd",
        "psym",
        "nsym",
        "kind",
    ]
    databaseindex_reader = csv.DictReader(
        codecs.iterdecode(line_gen, "utf-8"), fieldnames=fieldnames
    )

    # Query the SuiteSparse Matrix collection
    if verbose:
        print("Querying SuiteSparse Collection.")
    matrix_attr_dict = None
    matid = 0
    for row in databaseindex_reader:
        matid += 1
        if row["group"] == group and row["name"] == name:
            matrix_attr_dict = row
            matrix_attr_dict["matid"] = f"{matid}"

    if matrix_attr_dict is None:
        raise ValueError(
            f"Could not find matrix '{name}' in the SuiteSparse database index."
        )

    # Create a SuiteSparseMatrix (and save to file)
    return SuiteSparseMatrix.from_database_entry(matrix_attr_dict)


class SuiteSparseMatrix(linops.Matrix):
    """SuiteSparse Matrix.

    Sparse matrix from the `SuiteSparse Matrix Collection
    <https://sparse.tamu.edu/>`_. [1]_ [2]_

    Parameters
    ----------
    matid :
        Unique identifier for the matrix in the database.
    group :
        Group this matrix belongs to.
    name :
        Name of this matrix.
    nnz  :
        Number of non-zero elements.
    is2d3d:
        Does this matrix come from a 2D or 3D discretization?
    isspd :
        Is this matrix symmetric, positive definite?
    psym :
        Degree of symmetry of the matrix pattern.
    nsym :
        Degree of numerical symmetry of the matrix.
    kind  :
        Information of the problem domain this matrix arises from.

    References
    ----------
    .. [1] Davis, TA and Hu, Y. The University of Florida sparse matrix
           collection. *ACM Transactions on Mathematical Software (TOMS)* 38.1 (
           2011): 1-25.
    .. [2] Kolodziej, Scott P., et al. The SuiteSparse matrix collection website
           interface. *Journal of Open Source Software* 4.35 (2019): 1244.
    """

    # pylint: disable="too-many-instance-attributes,too-many-arguments,abstract-method"
    def __init__(
        self,
        matid: str,
        group: str,
        name: str,
        nnz: int,
        is2d3d: bool,
        isspd: bool,
        psym: float,
        nsym: float,
        kind: str,
    ):
        self.matid = matid
        self.group = group
        self.name = name
        self.nnz = nnz
        self.is2d3d = is2d3d
        self.isspd = isspd
        self.psym = psym
        self.nsym = nsym
        self.kind = kind

        super().__init__(A=self._download())

    @classmethod
    def from_database_entry(cls, database_entry: Dict) -> "SuiteSparseMatrix":
        """Create a SuiteSparseMatrix object from an entry of the database index.

        Parameters
        ----------
        database_entry :
            Dictionary representing one entry from the SuiteSparse database index.
        """
        return cls(
            matid=database_entry["matid"],
            group=database_entry["group"],
            nnz=int(database_entry["nnz"]),
            is2d3d=bool(int(database_entry["is2d3d"])),
            isspd=bool(int(database_entry["isspd"])),
            psym=float(database_entry["psym"]),
            nsym=float(database_entry["nsym"]),
            name=database_entry["name"],
            kind=database_entry["kind"],
        )

    def _download(
        self, verbose: bool = False
    ) -> Union[np.ndarray, scipy.sparse.coo_matrix]:
        """Download and extract file archive containing the sparse matrix.

        verbose:
            Print additional information.
        """
        try:
            import requests  # pylint: disable=import-outside-toplevel
        except ImportError as err:
            raise ImportError(
                "Cannot query SuiteSparse Matrix collection without optional dependency `requests`. Install ProbNum with optional dependencies for the problem zoo via `pip install probnum[zoo]` or install requests directly: `pip install requests`."
            ) from err

        url = SUITESPARSE_ROOT_URL + f"/MM/{self.group}/{self.name}.tar.gz"
        response = requests.get(url, stream=True)

        # Write archive to temporary file
        if verbose:
            print("Downloading compressed matrix.")

        chunk_size = 4096
        chunk_iter = response.iter_content(chunk_size=chunk_size)
        buffer = io.BytesIO()

        try:
            from tqdm.auto import tqdm  # pylint: disable=import-outside-toplevel

            with tqdm(
                total=int(response.headers["content-length"]),
                desc=self.name,
                unit="B",
                unit_scale=True,
            ) as pbar:
                for chunk in chunk_iter:
                    buffer.write(chunk)
                    pbar.update(chunk_size)
        except ImportError:
            for chunk in chunk_iter:
                buffer.write(chunk)

        buffer.seek(0)
        if verbose:
            print("Extracting file archive.")
        tar = tarfile.open(fileobj=buffer, mode="r:gz")

        return scipy.io.mmread(tar.extractfile(tar.getmembers()[0]))

    @staticmethod
    def _html_header() -> str:
        """Header of the HTML representation of a SuiteSparseMatrix."""
        return (
            "<thead>"
            + "".join(
                f"<th>{attr}</th>"
                for attr in [
                    "ID",
                    "Group",
                    "Name",
                    "Rows",
                    "Cols",
                    "Nonzeros",
                    "DType",
                    "2D/3D Discretization",
                    "SPD",
                    "Pattern Symmetry",
                    "Numerical Symmetry",
                    "Domain",
                    "Preview",
                ]
            )
            + "</thead>"
        )

    def _to_html_row(self) -> str:
        return (
            "<tr>"
            + "".join(
                f"<td>{str(table_item)}</td>"
                for table_item in [
                    self.matid,
                    (
                        f'<a href="{SUITESPARSE_ROOT_URL + "/" + self.group}"'
                        f' target="_blank">{self.group}</a>'
                    ),
                    (
                        f'<a href="'
                        f'{SUITESPARSE_ROOT_URL + "/" + self.group + "/" + self.name}" '
                        f'target="_blank">{self.name}</a>'
                    ),
                    self.shape[0],
                    self.shape[1],
                    self.nnz,
                    self.dtype,
                    self.is2d3d,
                    self.isspd,
                    f"{self.psym:.2}",
                    f"{self.nsym:.2}",
                    self.kind,
                    f'<img src="{SUITESPARSE_ROOT_URL}/files/{self.group}/{self.name}.png">',
                ]
            )
            + "</tr>"
        )

    def _repr_html_(self) -> str:
        """HTML representation."""
        return (
            f"<table>{self._html_header()}"
            + f"<tbody>{self._to_html_row()}</tbody></table>"
        )



"""Bayesian Quadrature.

This module provides routines to integrate functions through Bayesian
quadrature, meaning a model over the integrand is constructed in order
to actively select evaluation points of the integrand to estimate the
value of the integral. Bayesian quadrature methods return a random
variable, specifying the belief about the true value of the integral.
"""

from typing import Callable, Dict, Optional, Tuple, Union

import numpy as np

from probnum.kernels import Kernel
from probnum.randvars import Normal
from probnum.type import FloatArgType

from ._integration_measures import IntegrationMeasure, LebesgueMeasure
from .bq_methods import BayesianQuadrature


# pylint: disable=too-many-arguments
def bayesquad(
    fun: Callable,
    input_dim: int,
    kernel: Optional[Kernel] = None,
    domain: Optional[
        Tuple[Union[np.ndarray, FloatArgType], Union[np.ndarray, FloatArgType]]
    ] = None,
    nevals: int = None,
    measure: Optional[IntegrationMeasure] = None,
    method: str = "vanilla",
    policy: str = "bmc",
) -> Tuple[Normal, Dict]:
    r"""Infer the solution of the uni- or multivariate integral :math:`\int_\Omega f(x) d \mu(x)`
    on a hyper-rectangle :math:`\Omega = [a_1, b_1] \times \cdots \times [a_D, b_D]`.

    Bayesian quadrature (BQ) infers integrals of the form

    .. math:: F = \int_\Omega f(x) d \mu(x),

    of a function :math:`f:\mathbb{R}^D \mapsto \mathbb{R}` integrated on the domain
    :math:`\Omega \subset \mathbb{R}^D` against a measure :math:`\mu: \mathbb{R}^D
    \mapsto \mathbb{R}`.

    Bayesian quadrature methods return a probability distribution over the solution :math:`F` with
    uncertainty arising from finite computation (here a finite number of function evaluations).
    They start out with a random process encoding the prior belief about the function :math:`f`
    to be integrated. Conditioned on either existing or acquired function evaluations according to a
    policy, they update the belief on :math:`f`, which is translated into a posterior measure over
    the integral :math:`F`.
    See Briol et al. [1]_ for a review on Bayesian quadrature.

    Parameters
    ----------
    fun :
        Function to be integrated.
    input_dim:
        Input dimension of the integration problem
    kernel:
        the kernel used for the GP model
    domain :
        *shape=(dim,)* -- Domain of integration. Contains lower and upper bound as int or ndarray.
    measure:
        Integration measure, defaults to the Lebesgue measure.
    nevals :
        Number of function evaluations.
    method :
        Type of Bayesian quadrature to use. The available options are

        ====================  ===========
         vanilla              ``vanilla``
         WSABI                ``wsabi``
        ====================  ===========

    policy :
        Type of acquisition strategy to use. Options are

        =======================  =======
         Bayesian Monte Carlo    ``bmc``
         Uncertainty Sampling    ``us``
         Mutual Information      ``mi``
         Integral Variance       ``iv``
        =======================  =======

    Returns
    -------
    integral :
        The integral of ``func`` on the domain.
    info :
        Information on the performance of the method.

    References
    ----------
    .. [1] Briol, F.-X., et al., Probabilistic integration: A role in statistical computation?,
       *Statistical Science 34.1*, 2019, 1-22, 2019

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(0)
    >>> input_dim = 1
    >>> domain = (0, 1)
    >>> def f(x):
    ...     return x
    >>> F, info = bayesquad(fun=f, input_dim=input_dim, domain=domain)
    >>> print(F.mean)
    0.5000
    """
    if domain is None and measure is None:
        raise ValueError(
            "You need to either specify an integration domain or an integration "
            "measure. The Lebesgue measure can only operate on a finite domain."
        )

    # Integration measure
    if measure is None:
        measure = LebesgueMeasure(domain=domain, dim=input_dim)

    # Choose Method
    bq_method = BayesianQuadrature.from_interface(
        input_dim=input_dim, kernel=kernel, method=method, policy=policy
    )

    if nevals is None:
        nevals = input_dim * 25

    # Integrate
    integral, info = bq_method.integrate(fun=fun, measure=measure, nevals=nevals)

    return integral, info



"""Definitions of problems currently solved by probabilistic numerical methods."""

import dataclasses
import typing

import numpy as np
import scipy.sparse

import probnum.linops as pnlo
import probnum.type as pntp
from probnum import randvars


@dataclasses.dataclass
class RegressionProblem:
    r"""Regression problem.

    Fit a stochastic process to data, given a likelihood (realised by a :obj:`DiscreteGaussian` transition).
    Solved by filters and smoothers in :mod:`probnum.filtsmooth`.

    Parameters
    ----------
    observations
        Observations of the latent process.
    locations
        Grid-points on which the observations were taken.
    solution
        Array containing solution to the problem at ``locations``. Used for testing and benchmarking.

    Examples
    --------
    >>> obs = [11.4123, -15.5123]
    >>> loc = [0.1, 0.2]
    >>> rp = RegressionProblem(observations=obs, locations=loc)
    >>> rp
    RegressionProblem(observations=[11.4123, -15.5123], locations=[0.1, 0.2], solution=None)
    >>> rp.observations
    [11.4123, -15.5123]
    """

    observations: np.ndarray
    locations: np.ndarray

    # Optional: ground truth for testing and benchmarking
    solution: typing.Optional[
        typing.Callable[[np.ndarray], typing.Union[float, np.ndarray]]
    ] = None


@dataclasses.dataclass
class InitialValueProblem:
    r"""First order ODE initial value problem.

    Compute a function :math:`y=y(t)` that solves

    .. math::
        \dot y(t) = f(t, y(t)), \quad y(t_0) = y_0

    on time-interval :math:`[t_0, t_\text{max}]`.
    Solved by probabilistic ODE solvers in :mod:`probnum.diffeq`.


    Parameters
    ----------
    f
        ODE vector-field.
    t0
        Initial point in time.
    tmax
        Final point in time.
    y0
        Initial value of the solution.
    df
        Jacobian of the ODE vector-field :math:`f=f(t,y)` with respect to the :math:`y` variable.
    ddf
        Hessian of the ODE vector-field :math:`f=f(t,y)` with respect to the :math:`y` variable.
    solution
        Closed form, analytic solution to the problem. Used for testing and benchmarking.
    dy0_all
        All initial derivatives up to some order.

    Examples
    --------
    >>> import numpy as np
    >>> def f(t, x):
    ...     return x*(1-x)
    >>> ivp = InitialValueProblem(f, t0=0., tmax=3., y0=0.1)
    >>> ivp.t0, ivp.tmax, ivp.y0
    (0.0, 3.0, 0.1)
    >>> np.round(ivp.f(ivp.t0, ivp.y0), 2)
    0.09
    """

    f: typing.Callable[[float, np.ndarray], np.ndarray]
    t0: float
    tmax: float
    y0: typing.Union[pntp.FloatArgType, np.ndarray]
    df: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None
    ddf: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None

    # For testing and benchmarking
    solution: typing.Optional[typing.Callable[[float, np.ndarray], np.ndarray]] = None


@dataclasses.dataclass
class LinearSystem:
    r"""Linear system of equations.

    Compute :math:`x` from :math:`Ax=b`.
    Solved by probabilistic linear solvers in :mod:`probnum.linalg`

    Parameters
    ----------
    A
        System matrix or linear operator.
    b
        Right-hand side vector or matrix.
    solution
        True solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> import numpy as np
    >>> A = np.eye(3)
    >>> b = np.arange(3)
    >>> lin_sys = LinearSystem(A, b)
    >>> lin_sys
    LinearSystem(A=array([[1., 0., 0.],
           [0., 1., 0.],
           [0., 0., 1.]]), b=array([0, 1, 2]), solution=None)
    """

    A: typing.Union[
        np.ndarray,
        scipy.sparse.spmatrix,
        pnlo.LinearOperator,
        randvars.RandomVariable,
    ]
    b: typing.Union[np.ndarray, randvars.RandomVariable]

    # For testing and benchmarking
    solution: typing.Optional[typing.Union[np.ndarray, randvars.RandomVariable]] = None


@dataclasses.dataclass
class QuadratureProblem:
    r"""Numerical computation of an integral.

    Compute the integral

        .. math::
            \int_\Omega f(x) \, \text{d} \mu(x)

    for a function :math:`f: \Omega \rightarrow \mathbb{R}`.
    For the time being, :math:`\mu` is the Lebesgue measure.
    Solved by quadrature rules in :mod:`probnum.quad`.

    Parameters
    ----------
    integrand
        Function to be integrated.
    lower_bd
        A number or a vector representing the lower bounds of the integrals.
    upper_bd
        A number or a vector representing the upper bounds of the integrals.
    output_dim
        Output dimension of the integrand.
    solution
        Closed form, analytic solution to the problem. Used for testing and benchmarking.

    Examples
    --------
    >>> import numpy as np
    >>> def integrand(x):
    ...     return np.linalg.norm(x)**2
    >>> lower_bd = 0.41
    >>> upper_bd = 4.32
    >>> qp1d = QuadratureProblem(integrand, lower_bd=lower_bd, upper_bd=upper_bd)
    >>> np.round(qp1d.integrand(0.2), 2)
    0.04
    >>> qp1d.lower_bd
    0.41
    >>>
    >>> lower_bd = [0., 0.]
    >>> upper_bd = [1., 1.]
    >>> qp2d = QuadratureProblem(integrand, lower_bd=lower_bd, upper_bd=upper_bd)
    >>> qp2d.upper_bd
    [1.0, 1.0]
    """

    integrand: typing.Callable[[np.ndarray], typing.Union[float, np.ndarray]]
    lower_bd: typing.Union[pntp.FloatArgType, np.ndarray]
    upper_bd: typing.Union[pntp.FloatArgType, np.ndarray]
    output_dim: typing.Optional[int] = 1

    # For testing and benchmarking
    solution: typing.Optional[
        typing.Union[float, np.ndarray, randvars.RandomVariable]
    ] = None



import numpy as np
import pytest

import probnum.problems.zoo.filtsmooth as filtsmooth_zoo
from probnum import problems

all_filtmooth_setups = pytest.mark.parametrize(
    "filtsmooth_setup",
    [
        filtsmooth_zoo.benes_daum(),
        filtsmooth_zoo.car_tracking(),
        filtsmooth_zoo.logistic_ode(),
        filtsmooth_zoo.ornstein_uhlenbeck(),
        filtsmooth_zoo.pendulum(),
    ],
)


@all_filtmooth_setups
def test_types(filtsmooth_setup):
    regression_problem, statespace_components = filtsmooth_setup
    assert isinstance(regression_problem, problems.RegressionProblem)
    assert isinstance(regression_problem.observations, np.ndarray)
    assert isinstance(regression_problem.locations, np.ndarray)
    assert isinstance(regression_problem.solution, np.ndarray)
    assert isinstance(statespace_components, dict)
    assert "dynamics_model" in statespace_components
    assert "measurement_model" in statespace_components
    assert "initrv" in statespace_components


@all_filtmooth_setups
def test_shapes(filtsmooth_setup):
    regression_problem, statespace_components = filtsmooth_setup
    assert regression_problem.locations.size == regression_problem.observations.shape[0]

    measurement = statespace_components["measurement_model"].forward_realization(
        statespace_components["initrv"].mean, t=1.0
    )[0]
    assert measurement.shape == regression_problem.observations[0].shape



"""Tests for SuiteSparse matrices and related functions."""

import numpy as np
import scipy.sparse

from probnum.problems.zoo.linalg import SuiteSparseMatrix


def test_downloaded_matrix_is_sparse(suitesparse_mat: SuiteSparseMatrix):
    """Test whether a sparse scipy matrix is returned."""
    assert isinstance(suitesparse_mat.A, scipy.sparse.spmatrix)


def test_attribute_parsing(suitesparse_mycielskian: SuiteSparseMatrix):
    """Test whether the attributes listed on the SuiteSparse Matrix Collection site are
    parsed correctly."""
    assert suitesparse_mycielskian.matid == "2759"
    assert suitesparse_mycielskian.group == "Mycielski"
    assert suitesparse_mycielskian.name == "mycielskian3"
    assert suitesparse_mycielskian.shape == (5, 5)
    assert suitesparse_mycielskian.nnz == 10
    assert not suitesparse_mycielskian.isspd
    assert suitesparse_mycielskian.psym == 1.0
    assert suitesparse_mycielskian.nsym == 1.0


def test_html_representation_returns_string(suitesparse_mat: SuiteSparseMatrix):
    """Test whether the HTML representation of a SuiteSparse Matrix returns a string."""
    assert isinstance(suitesparse_mat._repr_html_(), str)


def test_trace(suitesparse_mycielskian: SuiteSparseMatrix):
    """Test whether the SuiteSparse Matrix object computes the trace correctly."""
    assert suitesparse_mycielskian.trace() == np.trace(
        suitesparse_mycielskian.todense()
    )


def test_matrix_multiplication(suitesparse_mat: SuiteSparseMatrix):
    """Test whether the SuiteSparse matrix can be multiplied with."""
    zerovec = np.zeros(suitesparse_mat.shape[1])
    identity = np.eye(suitesparse_mat.shape[1])
    np.testing.assert_allclose(
        np.zeros(suitesparse_mat.shape[0]), suitesparse_mat @ zerovec
    )
    np.testing.assert_allclose(suitesparse_mat.todense(), suitesparse_mat @ identity)



"""Random Variables."""

from ._categorical import Categorical
from ._constant import Constant
from ._normal import Normal
from ._random_variable import (
    ContinuousRandomVariable,
    DiscreteRandomVariable,
    RandomVariable,
)
from ._scipy_stats import (
    WrappedSciPyContinuousRandomVariable,
    WrappedSciPyDiscreteRandomVariable,
    WrappedSciPyRandomVariable,
)
from ._utils import asrandvar

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "asrandvar",
    "RandomVariable",
    "DiscreteRandomVariable",
    "ContinuousRandomVariable",
    "Constant",
    "Normal",
    "Categorical",
    "WrappedSciPyRandomVariable",
    "WrappedSciPyDiscreteRandomVariable",
    "WrappedSciPyContinuousRandomVariable",
]

# Set correct module paths. Corrects links and module paths in documentation.
RandomVariable.__module__ = "probnum.randvars"
DiscreteRandomVariable.__module__ = "probnum.randvars"
ContinuousRandomVariable.__module__ = "probnum.randvars"

WrappedSciPyRandomVariable.__module__ = "probnum.randvars"
WrappedSciPyDiscreteRandomVariable.__module__ = "probnum.randvars"
WrappedSciPyContinuousRandomVariable.__module__ = "probnum.randvars"

Constant.__module__ = "probnum.randvars"
Normal.__module__ = "probnum.randvars"
Categorical.__module__ = "probnum.randvars"

asrandvar.__module__ = "probnum.randvars"



import unittest

import probnum.filtsmooth as pnfs
import probnum.statespace as pnss

from .. import filtsmooth_testcases as cases


class TestContinuousUKFComponent(unittest.TestCase):
    """Implementation incomplete, hence check that an error is raised."""

    def test_notimplementederror(self):
        sde = pnss.SDE(1, None, None, None)  # content is irrelevant.
        with self.assertRaises(NotImplementedError):
            pnfs.ContinuousUKFComponent(sde)


class TestDiscreteUKFComponent(cases.LinearisedDiscreteTransitionTestCase):
    def setUp(self):
        self.linearising_component_pendulum = pnfs.DiscreteUKFComponent
        self.linearising_component_car = pnfs.DiscreteUKFComponent
        self.visualise = False



"""Bayesian Quadrature."""

from ._bayesquad import bayesquad
from ._integration_measures import GaussianMeasure, IntegrationMeasure, LebesgueMeasure
from .bq_methods import BayesianQuadrature
from .kernel_embeddings import (
    KernelEmbedding,
    _kernel_mean_expquad_gauss,
    _kernel_mean_expquad_lebesgue,
    _kernel_variance_expquad_gauss,
    _kernel_variance_expquad_lebesgue,
)
from .policies import sample_from_measure

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "bayesquad",
    "BayesianQuadrature",
    "IntegrationMeasure",
    "KernelEmbedding",
    "GaussianMeasure",
    "LebesgueMeasure",
]

# Set correct module paths. Corrects links and module paths in documentation.
BayesianQuadrature.__module__ = "probnum.quad"
IntegrationMeasure.__module__ = "probnum.quad"
KernelEmbedding.__module__ = "probnum.quad"
GaussianMeasure.__module__ = "probnum.quad"
LebesgueMeasure.__module__ = "probnum.quad"



"""Test cases for kernel embeddings."""

from typing import Optional, Tuple, Union

import numpy as np
import pytest
from scipy.linalg import sqrtm
from scipy.special import roots_legendre

from probnum.type import FloatArgType, IntArgType


# Auxiliary functions
def gauss_hermite_tensor(
    n_points: IntArgType,
    dim: IntArgType,
    mean: Union[np.ndarray, FloatArgType],
    cov: Union[np.ndarray, FloatArgType],
):
    """Returns the points and weights of a tensor-product Gauss-Hermite rule for
    integration w.r.t a Gaussian measure."""
    x_gh_1d, w_gh = np.polynomial.hermite.hermgauss(n_points)
    x_gh = (
        np.sqrt(2)
        * np.stack(np.meshgrid(*(x_gh_1d,) * dim), -1).reshape(-1, dim)
        @ sqrtm(np.atleast_2d(cov))
        + mean
    )
    w_gh = np.prod(
        np.stack(np.meshgrid(*(w_gh,) * dim), -1).reshape(-1, dim), axis=1
    ) / (np.pi ** (dim / 2))
    return x_gh, w_gh


def gauss_legendre_tensor(
    n_points: IntArgType,
    dim: IntArgType,
    domain: Tuple[Union[np.ndarray, FloatArgType], Union[np.ndarray, FloatArgType]],
    normalized: Optional[bool] = False,
):
    """Returns the points and weights of a tensor-product Gauss-Legendre rule for
    integration w.r.t the Lebesgue measure on a hyper-rectangle."""
    x_1d, w_gl, weight_sum = roots_legendre(n_points, True)
    x_1d_shifted = [
        0.5 * (x_1d * (domain[1][i] - domain[0][i]) + domain[1][i] + domain[0][i])
        for i in range(0, dim)
    ]
    x_gl = np.stack(np.meshgrid(*x_1d_shifted), -1).reshape(-1, dim)
    w_gl = (
        np.prod(np.stack(np.meshgrid(*(w_gl,) * dim), -1).reshape(-1, dim), axis=1)
        / weight_sum ** dim
    )
    if not normalized:
        w_gl = w_gl * np.prod(domain[1] - domain[0])
    return x_gl, w_gl


# Common tests
def test_kernel_mean_shape(kernel_embedding, x):
    """Test output shape of kernel mean."""

    kernel_mean_shape = (np.atleast_2d(x).shape[0],)

    assert kernel_embedding.kernel_mean(x).shape == kernel_mean_shape, (
        f"Kernel mean of {type(kernel_embedding)} has shape"
        f" {kernel_embedding.kernel_mean(x).shape} instead of {kernel_mean_shape}"
    )


def test_kernel_variance_float(kernel_embedding):
    """Test output of kernel variance."""
    assert isinstance(kernel_embedding.kernel_variance(), float)


# Tests for squared exponential kernel and Gaussian measure
@pytest.mark.parametrize("input_dim", [1, 2, 3, 5])
@pytest.mark.parametrize("measure_name", ["gauss"])
def test_kernel_mean_gaussian_measure(kernel_embedding, num_data):
    """Test kernel means for the Gaussian measure against Gauss-Hermite tensor product
    rule."""
    n_gh = 10
    x_gh, w_gh = gauss_hermite_tensor(
        n_points=n_gh,
        dim=kernel_embedding.dim,
        mean=kernel_embedding.measure.mean,
        cov=kernel_embedding.measure.cov,
    )

    x = kernel_embedding.measure.sample(num_data)
    num_kernel_means = kernel_embedding.kernel(x, x_gh) @ w_gh
    true_kernel_means = kernel_embedding.kernel_mean(x)
    np.testing.assert_allclose(
        true_kernel_means, num_kernel_means, rtol=1.0e-2, atol=1.0e-2
    )


@pytest.mark.parametrize("input_dim", [1, 2, 3, 5])
@pytest.mark.parametrize("measure_name", ["gauss"])
def test_kernel_var_gaussian_measure(kernel_embedding):
    """Test kernel variance for the Gaussian measure against Gauss-Hermite tensor
    product rule."""
    n_gh = 20
    x_gh, w_gh = gauss_hermite_tensor(
        n_points=n_gh,
        dim=kernel_embedding.dim,
        mean=kernel_embedding.measure.mean,
        cov=kernel_embedding.measure.cov,
    )

    num_kernel_variance = kernel_embedding.kernel_mean(x_gh) @ w_gh
    true_kernel_variance = kernel_embedding.kernel_variance()
    np.testing.assert_allclose(
        true_kernel_variance, num_kernel_variance, rtol=1.0e-3, atol=1.0e-3
    )


# Tests for squared exponential kernel and Lebesgue measure
@pytest.mark.parametrize("input_dim", [1, 2, 3, 5])
@pytest.mark.parametrize("measure_name", ["lebesgue"])
def test_kernel_mean_lebesgue_measure(kernel_embedding, num_data):
    """Test kernel means for the Lebesgue measure against Gauss-Legendre tensor product
    rule."""
    n_gl = 10
    x_gl, w_gl = gauss_legendre_tensor(
        n_points=n_gl,
        dim=kernel_embedding.dim,
        domain=kernel_embedding.measure.domain,
        normalized=kernel_embedding.measure.normalized,
    )

    x = kernel_embedding.measure.sample(num_data)
    num_kernel_means = kernel_embedding.kernel(x, x_gl) @ w_gl
    true_kernel_means = kernel_embedding.kernel_mean(x)
    np.testing.assert_allclose(
        true_kernel_means, num_kernel_means, rtol=1.0e-3, atol=1.0e-3
    )


@pytest.mark.parametrize("input_dim", [1, 2, 3, 5])
@pytest.mark.parametrize("measure_name", ["lebesgue"])
def test_kernel_var_lebesgue_measure(kernel_embedding):
    """Test kernel variance for the Lebesgue measure against Gauss-Legendre tensor
    product rule."""
    n_gl = 20
    x_gl, w_gl = gauss_legendre_tensor(
        n_points=n_gl,
        dim=kernel_embedding.dim,
        domain=kernel_embedding.measure.domain,
        normalized=kernel_embedding.measure.normalized,
    )
    num_kernel_variance = kernel_embedding.kernel_mean(x_gl) @ w_gl
    true_kernel_variance = kernel_embedding.kernel_variance()
    np.testing.assert_allclose(
        true_kernel_variance, num_kernel_variance, rtol=1.0e-3, atol=1.0e-3
    )



"""Cholesky updates."""


import typing

import numpy as np

__all__ = ["cholesky_update", "tril_to_positive_tril"]


def cholesky_update(
    S1: np.ndarray, S2: typing.Optional[np.ndarray] = None
) -> np.ndarray:
    r"""Compute Cholesky update/factorization :math:`L` such that :math:`L L^\top = S_1 S_1^\top + S_2 S_2^\top` holds.

    This can be used in various ways.
    For example, :math:`S_1` and :math:`S_2` do not need to be Cholesky factors; any matrix square-root is sufficient.
    As long as :math:`L L^\top = S_1 S_1^\top + S_2 S_2^\top` is well-defined (and admits a Cholesky-decomposition),
    :math:`S_1` and :math:`S_2` do not even have to be square.


    Parameters
    ----------
    S1 :
        First matrix square-root. Not necessarily a Cholesky factor, any (possibly even non-square) matrix :math:`S` such that :math:`C = S S^\top` holds, is sufficient.
    S2 :
        Second matrix square-root. Not necessarily a Cholesky factor, any (possibly even non-square) matrix :math:`S` such that :math:`C = S S^\top` holds, is sufficient.
        Optional. Default is None.

    Returns
    -------
    Lower Cholesky factor :math:`L` of :math:`L L^\top =S1 S1^\top + S2 S2^\top`, if ``S2`` was not None. Otherwise, lower Cholesky factor of :math:`L L^\top =S1 S1^\top`.


    Examples
    --------

    >>> from probnum.utils.linalg import cholesky_update
    >>> from probnum.problems.zoo.linalg import random_spd_matrix

    Compute the Cholesky-factor of a sum of SPD matrices.

    >>> C1 = random_spd_matrix(5)
    >>> S1 = np.linalg.cholesky(C1)
    >>> C2 = random_spd_matrix(5)
    >>> S2 = np.linalg.cholesky(C2)
    >>> C = C1 + C2
    >>> S = cholesky_update(S1, S2)
    >>> np.allclose(np.linalg.cholesky(C), S)
    True

    Turn a (potentially non-square) matrix square-root into a Cholesky factor

    >>> A = np.random.rand(3, 5)
    >>> S = cholesky_update(A @ S1)
    >>> np.allclose(np.linalg.cholesky(A @ C1 @ A.T), S)
    True
    """
    if S2 is not None:
        stacked_up = np.vstack((S1.T, S2.T))
    else:
        stacked_up = np.vstack(S1.T)
    upper_sqrtm = np.linalg.qr(stacked_up, mode="r")
    lower_sqrtm = upper_sqrtm.T
    return tril_to_positive_tril(lower_sqrtm)


def tril_to_positive_tril(tril_mat: np.ndarray) -> np.ndarray:
    r"""Orthogonally transform a lower-triangular matrix into a lower-triangular matrix with positive diagonal.

    In other words, make it a valid lower Cholesky factor.

    The name of the function is based on `np.tril`.
    """
    d = np.sign(np.diag(tril_mat))

    # Numpy assigns sign 0 to 0.0, which eliminate entire rows in the operation below.
    d[d == 0] = 1.0

    # Fast(er) multiplication with a diagonal matrix from the right via broadcasting.
    with_pos_diag = tril_mat * d[None, :]
    return with_pos_diag



"""Test problems from linear algebra."""

from ._random_spd_matrix import random_sparse_spd_matrix, random_spd_matrix
from ._suitesparse_matrix import SuiteSparseMatrix, suitesparse_matrix

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "random_spd_matrix",
    "random_sparse_spd_matrix",
    "suitesparse_matrix",
    "SuiteSparseMatrix",
]

# Set correct module paths. Corrects links and module paths in documentation.
SuiteSparseMatrix.__module__ = "probnum.problems.zoo.linalg"



import numpy as np
import pytest

import probnum.problems.zoo.filtsmooth as filtsmooth_zoo
from probnum import filtsmooth


@pytest.fixture(params=[filtsmooth_zoo.logistic_ode])
def setup(request):
    """Filter and regression problem."""
    problem = request.param
    regression_problem, statespace_components = problem()

    kalman = filtsmooth.Kalman(
        statespace_components["dynamics_model"],
        statespace_components["measurement_model"],
        statespace_components["initrv"],
    )
    return kalman, regression_problem


def test_rmse_filt_smooth(setup):
    """Assert that iterated smoothing beats smoothing."""

    np.random.seed(12345)
    kalman, regression_problem = setup
    truth = regression_problem.solution

    stopcrit = filtsmooth.StoppingCriterion(atol=1e-1, rtol=1e-1, maxit=10)

    posterior, _ = kalman.filter(regression_problem)
    posterior = kalman.smooth(posterior)

    iterated_posterior, _ = kalman.iterated_filtsmooth(
        regression_problem, stopcrit=stopcrit
    )

    filtms = posterior.filtering_posterior.states.mean
    smooms = posterior.states.mean
    iterms = iterated_posterior.states.mean

    if filtms.ndim == 1:
        filtms = filtms.reshape((-1, 1))
        smooms = smooms.reshape((-1, 1))
        iterms = iterms.reshape((-1, 1))

    if truth.ndim == 1:
        truth = truth.reshape((-1, 1))

    # Compare only zeroth component
    # for compatibility with all test cases
    smooms_rmse = np.mean(np.abs(smooms[:, 0] - truth[:, 0]))
    iterms_rmse = np.mean(np.abs(iterms[:, 0] - truth[:, 0]))

    assert iterms_rmse < smooms_rmse



"""ODE solutions returned by Gaussian ODE filtering."""

from typing import Optional

import numpy as np

from probnum import _randomvariablelist, filtsmooth, randvars, utils
from probnum.filtsmooth.timeseriesposterior import DenseOutputLocationArgType
from probnum.type import FloatArgType, RandomStateArgType, ShapeArgType

from ..odesolution import ODESolution


class KalmanODESolution(ODESolution):
    """Gaussian IVP filtering solution of an ODE problem.

    Recall that in ProbNum, Gaussian filtering and smoothing is generally named "Kalman".

    Parameters
    ----------
    kalman_posterior
        Gauss-Markov posterior over the ODE solver state space model.
        Therefore, it assumes that the dynamics model is an :class:`Integrator`.

    See Also
    --------
    GaussianIVPFilter : ODE solver that behaves like a Gaussian filter.
    KalmanPosterior : Posterior over states after Gaussian filtering/smoothing.

    Examples
    --------
    >>> from probnum.diffeq import logistic, probsolve_ivp
    >>> from probnum import randvars
    >>>
    >>> def f(t, x):
    ...     return 4*x*(1-x)
    >>>
    >>> y0 = np.array([0.15])
    >>> t0, tmax = 0., 1.5
    >>> solution = probsolve_ivp(f, t0, tmax, y0, step=0.1, adaptive=False)
    >>> # Mean of the discrete-time solution
    >>> print(np.round(solution.states.mean, 2))
    [[0.15]
     [0.21]
     [0.28]
     [0.37]
     [0.47]
     [0.57]
     [0.66]
     [0.74]
     [0.81]
     [0.87]
     [0.91]
     [0.94]
     [0.96]
     [0.97]
     [0.98]
     [0.99]]

    >>> # Times of the discrete-time solution
    >>> print(solution.locations)
    [0.  0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.  1.1 1.2 1.3 1.4 1.5]
    >>> # Individual entries of the discrete-time solution can be accessed with
    >>> print(solution[5])
    <Normal with shape=(1,), dtype=float64>
    >>> print(np.round(solution[5].mean, 2))
    [0.56]
    >>> # Evaluate the continuous-time solution at a new time point t=0.65
    >>> print(np.round(solution(0.65).mean, 2))
    [0.70]
    """

    def __init__(self, kalman_posterior: filtsmooth.KalmanPosterior):
        self.kalman_posterior = kalman_posterior

        # Pre-compute projection matrices.
        # The prior must be an integrator, if not, an error is thrown in 'GaussianIVPFilter'.
        self.proj_to_y = self.kalman_posterior.transition.proj2coord(coord=0)
        self.proj_to_dy = self.kalman_posterior.transition.proj2coord(coord=1)

        states = _randomvariablelist._RandomVariableList(
            [_project_rv(self.proj_to_y, rv) for rv in self.kalman_posterior.states]
        )
        derivatives = _randomvariablelist._RandomVariableList(
            [_project_rv(self.proj_to_dy, rv) for rv in self.kalman_posterior.states]
        )
        super().__init__(
            locations=kalman_posterior.locations, states=states, derivatives=derivatives
        )

    def interpolate(
        self,
        t: FloatArgType,
        previous_location: Optional[FloatArgType] = None,
        previous_state: Optional[randvars.RandomVariable] = None,
        next_location: Optional[FloatArgType] = None,
        next_state: Optional[randvars.RandomVariable] = None,
    ) -> randvars.RandomVariable:
        out_rv = self.kalman_posterior.interpolate(
            t, previous_location, previous_state, next_location, next_state
        )
        return _project_rv(self.proj_to_y, out_rv)

    def sample(
        self,
        t: Optional[DenseOutputLocationArgType] = None,
        size: Optional[ShapeArgType] = (),
        random_state: Optional[RandomStateArgType] = None,
    ) -> np.ndarray:

        samples = self.kalman_posterior.sample(
            t=t, size=size, random_state=random_state
        )
        # Project the samples down to the "true" KalmanODESolution dimensions
        # (which are a subset of the KalmanPosterior dimensions)
        ode_samples = np.einsum("dq,...q->...d", self.proj_to_y, samples)

        return ode_samples

    def transform_base_measure_realizations(
        self,
        base_measure_realizations: np.ndarray,
        t: DenseOutputLocationArgType = None,
    ) -> np.ndarray:
        errormsg = (
            "The KalmanODESolution does not implement transformation of realizations of a base measure."
            "Try `KalmanODESolution.kalman_posterior.transform_base_measure_realizations` instead."
        )

        raise NotImplementedError(errormsg)

    @property
    def filtering_solution(self):

        if isinstance(self.kalman_posterior, filtsmooth.FilteringPosterior):
            return self

        # else: self.kalman_posterior is a SmoothingPosterior object, which has the field filter_posterior.
        return KalmanODESolution(
            kalman_posterior=self.kalman_posterior.filtering_posterior
        )

    @property
    def _states_left_of_location(self):
        """Return the set of states that is used to find the LEFTMOST states of a given
        time point in a way that supports slicing with the output of numpy.searchsorted.

        Thus, the output is wrapped into a numpy array (which is not something we want all the time,
        because then the _RandomVariableList functionality would be lost.


        Note: This exists as a property, because for the KalmanSmoothingPosterior, the leftmost states
        are extracted from the filtering posterior, not the smoothing posterior, which can be overwritten here.
        """
        return self.kalman_posterior._states_left_of_location

    @property
    def _states_right_of_location(self):
        """Return the set of states that is used to find the RIGHTMOST states of a given
        time point in a way that supports slicing with the output of numpy.searchsorted.

        Thus, the output is wrapped into a numpy array (which is not something we want all the time,
        because then the _RandomVariableList functionality would be lost.

        This exists as a property because the leftmost states exist as a property and who knows when we need it like that.
        """
        return self.kalman_posterior._states_right_of_location


def _project_rv(projmat, rv):
    # There is no way of checking whether `rv` has its Cholesky factor computed already or not.
    # Therefore, since we need to update the Cholesky factor for square-root filtering,
    # we also update the Cholesky factor for non-square-root algorithms here,
    # which implies additional cost.
    # See Issues #319 and #329.
    # When they are resolved, this function here will hopefully be superfluous.

    new_mean = projmat @ rv.mean
    new_cov = projmat @ rv.cov @ projmat.T
    new_cov_cholesky = utils.linalg.cholesky_update(projmat @ rv.cov_cholesky)
    return randvars.Normal(new_mean, new_cov, cov_cholesky=new_cov_cholesky)



"""Gaussian filtering and smoothing."""

from .extendedkalman import ContinuousEKFComponent, DiscreteEKFComponent, EKFComponent
from .iterated_component import IteratedDiscreteComponent
from .kalman import Kalman
from .kalmanposterior import FilteringPosterior, KalmanPosterior, SmoothingPosterior
from .stoppingcriterion import StoppingCriterion
from .unscentedkalman import ContinuousUKFComponent, DiscreteUKFComponent, UKFComponent
from .unscentedtransform import UnscentedTransform



"""Abstract Base Class for posteriors over states after applying filtering/smoothing."""

import abc
from typing import Optional, Union

import numpy as np

from probnum import _randomvariablelist, randvars
from probnum.type import (
    ArrayLikeGetitemArgType,
    FloatArgType,
    RandomStateArgType,
    ShapeArgType,
)

DenseOutputLocationArgType = Union[FloatArgType, np.ndarray]
"""TimeSeriesPosteriors and derived classes can be evaluated at a single location 't'
or an array of locations."""

DenseOutputValueType = Union[
    randvars.RandomVariable, _randomvariablelist._RandomVariableList
]
"""Dense evaluation of a TimeSeriesPosterior returns a RandomVariable if evaluated at a single location,
and a _RandomVariableList if evaluated at an array of locations."""


class TimeSeriesPosterior(abc.ABC):
    """Posterior Distribution over States after time-series algorithms such as
    filtering/smoothing or solving ODEs.

    Parameters
    ----------
    locations :
        Locations of the posterior states (represented as random variables).
    states :
        Posterior random variables.
    """

    def __init__(self, locations: np.ndarray, states: np.ndarray) -> None:
        self.locations = np.asarray(locations)
        self.states = _randomvariablelist._RandomVariableList(states)

    def __len__(self) -> int:
        """Length of the discrete-time solution.

        Corresponds to the number of filtering/smoothing steps.
        """
        return len(self.locations)

    def __getitem__(self, idx: ArrayLikeGetitemArgType) -> randvars.RandomVariable:
        return self.states[idx]

    def __call__(self, t: DenseOutputLocationArgType) -> DenseOutputValueType:
        """Evaluate the time-continuous posterior at location `t`

        Algorithm:
        1. Find closest t_prev and t_next, with t_prev < t < t_next
        2. Predict from t_prev to t
        3. (if `self._with_smoothing=True`) Predict from t to t_next
        4. (if `self._with_smoothing=True`) Smooth from t_next to t
        5. Return random variable for time t

        Parameters
        ----------
        t :
            Location, or time, at which to evaluate the posterior.

        Returns
        -------
        randvars.RandomVariable or _randomvariablelist._RandomVariableList
            Estimate of the states at time ``t``.
        """
        # The variable "squeeze_eventually" indicates whether
        # the dimension of the time-array has been promoted
        # (which is there for a well-behaved loop).
        # If this has been the case, the final result needs to be
        # reshaped ("squeezed") accordingly.
        if np.isscalar(t):
            t = np.atleast_1d(t)
            squeeze_eventually = True
        else:
            squeeze_eventually = False

        if not np.all(np.diff(t) >= 0.0):
            raise ValueError("Time-points have to be sorted.")

        # Split left-extrapolation, interpolation, right_extrapolation
        t0, tmax = np.amin(self.locations), np.amax(self.locations)
        t_extra_left = t[t < t0]
        t_extra_right = t[t > tmax]
        t_inter = t[(t0 <= t) & (t <= tmax)]

        # Indices of t where they would be inserted
        # into self.locations ("left": right-closest states)
        indices = np.searchsorted(self.locations, t_inter, side="left")
        interpolated_values = [
            self.interpolate(
                t=ti,
                previous_location=prevloc,
                previous_state=prevstate,
                next_location=nextloc,
                next_state=nextstate,
            )
            for ti, prevloc, prevstate, nextloc, nextstate in zip(
                t_inter,
                self.locations[indices - 1],
                self._states_left_of_location[indices - 1],
                self.locations[indices],
                self._states_right_of_location[indices],
            )
        ]
        extrapolated_values_left = [
            self.interpolate(
                t=ti,
                previous_location=None,
                previous_state=None,
                next_location=t0,
                next_state=self._states_right_of_location[0],
            )
            for ti in t_extra_left
        ]

        extrapolated_values_right = [
            self.interpolate(
                t=ti,
                previous_location=tmax,
                previous_state=self._states_left_of_location[-1],
                next_location=None,
                next_state=None,
            )
            for ti in t_extra_right
        ]
        dense_output_values = extrapolated_values_left
        dense_output_values.extend(interpolated_values)
        dense_output_values.extend(extrapolated_values_right)

        if squeeze_eventually:
            return dense_output_values[0]
        return _randomvariablelist._RandomVariableList(dense_output_values)

    @abc.abstractmethod
    def interpolate(
        self,
        t: FloatArgType,
        previous_location: Optional[FloatArgType] = None,
        previous_state: Optional[randvars.RandomVariable] = None,
        next_location: Optional[FloatArgType] = None,
        next_state: Optional[randvars.RandomVariable] = None,
    ) -> randvars.RandomVariable:
        """Evaluate the posterior at a measurement-free point.

        Parameters
        ----------
        t :
            Location to evaluate at.

        Returns
        -------
        randvars.RandomVariable or _randomvariablelist._RandomVariableList
            Dense evaluation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def sample(
        self,
        t: Optional[DenseOutputLocationArgType] = None,
        size: Optional[ShapeArgType] = (),
        random_state: Optional[RandomStateArgType] = None,
    ) -> np.ndarray:
        """Draw samples from the filtering/smoothing posterior.

        If nothing is specified, a single sample is drawn (supported on self.locations).
        If locations are specified, a single sample is drawn on those locations.
        If size is specified, more than a single sample is drawn.

        Internally, samples from a base measure are drawn and transformed via self.transform_base_measure_realizations.

        Parameters
        ----------
        t :
            Locations on which the samples are wanted. Default is none, which implies that
            self.location is used.
        size :
            Indicates how many samples are drawn. Default is an empty tuple, in which case
            a single sample is returned.
        random_state
            Random state (seed, generator) to be used for sampling base measure realizations.


        Returns
        -------
        np.ndarray
            Drawn samples. If size has shape (A1, ..., Z1), locations have shape (L,),
            and the state space model has shape (A2, ..., Z2), the output has
            shape (A1, ..., Z1, L, A2, ..., Z2).
            For example: size=4, len(locations)=4, dim=3 gives shape (4, 4, 3).
        """
        raise NotImplementedError("Sampling is not implemented.")

    @abc.abstractmethod
    def transform_base_measure_realizations(
        self,
        base_measure_realizations: np.ndarray,
        t: Optional[DenseOutputLocationArgType],
    ) -> np.ndarray:
        """Transform a set of realizations from a base measure into realizations from
        the posterior.

        Parameters
        ----------
        base_measure_realizations :
            Base measure realizations.
        t :
            Locations on which the transformed realizations shall represent realizations from the posterior.

        Returns
        -------
        np.ndarray
            Transformed realizations.
        """
        raise NotImplementedError(
            "Transforming base measure realizations is not implemented."
        )

    @property
    def _states_left_of_location(self):
        """Return the set of states that is used to find the LEFTMOST states of a given
        time point in a way that supports slicing with the output of numpy.searchsorted.

        Thus, the output is wrapped into a numpy array (which is not something we want all the time,
        because then the _RandomVariableList functionality would be lost.


        Note: This exists as a property, because for the KalmanSmoothingPosterior, the leftmost states
        are extracted from the filtering posterior, not the smoothing posterior, which can be overwritten here.
        """
        return np.asarray(self.states)

    @property
    def _states_right_of_location(self):
        """Return the set of states that is used to find the RIGHTMOST states of a given
        time point in a way that supports slicing with the output of numpy.searchsorted.

        Thus, the output is wrapped into a numpy array (which is not something we want all the time,
        because then the _RandomVariableList functionality would be lost.

        This exists as a property because the leftmost states exist as a property and who knows when we need it like that.
        """
        return np.asarray(self.states)



"""Benchmarks for random variables."""

import numpy as np

import probnum.kernels as kernels

# Module level variables
KERNEL_NAMES = [
    "white_noise",
    "linear",
    "polynomial",
    "exp_quad",
    "rat_quad",
    "matern12",
    "matern32",
    "matern52",
    "matern72",
]

N_DATAPOINTS = [10, 100, 1000]


def get_kernel(kernel_name, input_dim):
    """Return a kernel for a given name."""
    if kernel_name == "white_noise":
        kernel = kernels.WhiteNoise(input_dim=input_dim)
    elif kernel_name == "linear":
        kernel = kernels.Linear(input_dim=input_dim)
    elif kernel_name == "polynomial":
        kernel = kernels.Polynomial(input_dim=input_dim)
    elif kernel_name == "exp_quad":
        kernel = kernels.ExpQuad(input_dim=input_dim)
    elif kernel_name == "rat_quad":
        kernel = kernels.RatQuad(input_dim=input_dim)
    elif kernel_name == "matern12":
        kernel = kernels.Matern(input_dim=input_dim, nu=0.5)
    elif kernel_name == "matern32":
        kernel = kernels.Matern(input_dim=input_dim, nu=1.5)
    elif kernel_name == "matern52":
        kernel = kernels.Matern(input_dim=input_dim, nu=2.5)
    elif kernel_name == "matern72":
        kernel = kernels.Matern(input_dim=input_dim, nu=3.5)
    else:
        raise ValueError(f"Kernel name '{kernel_name}' not recognized.")

    return kernel


class Kernels:
    """Benchmark evaluation of a kernel at a set of inputs."""

    param_names = ["kernel", "n_datapoints"]
    params = [KERNEL_NAMES, N_DATAPOINTS]

    def setup(self, kernel, n_datapoints):
        rng = np.random.default_rng(42)
        self.input_dim = 100
        self.data = rng.normal(size=(n_datapoints, self.input_dim))
        self.kernel = get_kernel(kernel_name=kernel, input_dim=self.input_dim)

    def time_kernel_matrix(self, kernel, n_datapoints):
        """Times sampling from this distribution."""
        self.kernel(self.data)

    def peakmem_kernel_matrix(self, kernel, n_datapoints):
        """Peak memory of sampling process."""
        self.kernel(self.data)



"""Stopping criteria for iterated filtering and smoothing."""
import numpy as np


class StoppingCriterion:
    """Stop iteration if absolute and relative tolerance are reached."""

    def __init__(self, atol=1e-3, rtol=1e-6, maxit=1000):
        self.atol = atol
        self.rtol = rtol
        self.maxit = maxit
        self.iterations = 0

    def terminate(self, error, reference):
        """Decide whether the stopping criterion is satisfied, which implies terminating
        of the iteration.

        If the error is sufficiently small (with respect to atol, rtol
        and the reference), return True. Else, return False. Throw a
        runtime error if the maximum number of iterations is reached.
        """
        if self.iterations > self.maxit:
            errormsg = f"Maximum number of iterations (N={self.maxit}) reached."
            raise RuntimeError(errormsg)
        magnitude = self.evaluate_error(error=error, reference=reference)
        if magnitude > 1:
            self.iterations += 1
            return False
        else:
            self.iterations = 0
            return True

    def evaluate_error(self, error, reference):
        """Compute the normalised error."""
        normalisation = self.atol + self.rtol * reference
        magnitude = np.sqrt(np.mean((error / normalisation) ** 2))
        return magnitude



"""Random Processes."""

import abc
from typing import Callable, Generic, Optional, Type, TypeVar, Union

import numpy as np

from probnum import _randomvariablelist, randvars
from probnum import utils as _utils
from probnum.type import DTypeArgType, IntArgType, RandomStateArgType, ShapeArgType

_InputType = TypeVar("InputType")
_OutputType = TypeVar("OutputType")


class RandomProcess(Generic[_InputType, _OutputType], abc.ABC):
    """Random processes represent uncertainty about a function.

    Random processes generalize functions by encoding uncertainty over function
    values in their covariance function. They can be used to model (deterministic)
    functions which are not fully known or to define functions with stochastic
    output.

    Parameters
    ----------
    input_dim :
        Input dimension of the random process.
    output_dim :
        Output dimension of the random process.
    dtype :
        Data type of the random process evaluated at an input. If ``object`` will be
        converted to ``numpy.dtype``.

    See Also
    --------
    RandomVariable : Random variables.
    GaussianProcess : Gaussian processes.
    MarkovProcess : Random processes with the Markov property.

    Notes
    -----
    Random processes are assumed to have an (un-/countably) infinite domain. Random
    processes with a finite index set are represented by :class:`RandomVariable`.
    """

    # pylint: disable=invalid-name

    def __init__(
        self,
        input_dim: IntArgType,
        output_dim: IntArgType,
        dtype: DTypeArgType,
    ):
        self._input_dim = np.int_(_utils.as_numpy_scalar(input_dim))
        self._output_dim = np.int_(_utils.as_numpy_scalar(output_dim))
        self._dtype = np.dtype(dtype)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} with "
            f"input_dim={self.input_dim}, output_dim={self.output_dim}, "
            f"dtype={self.dtype}>"
        )

    @abc.abstractmethod
    def __call__(self, args: _InputType) -> randvars.RandomVariable[_OutputType]:
        """Evaluate the random process at a set of input arguments.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Input(s) to evaluate random
            process at.

        Returns
        -------
        randvars.RandomVariable
            *shape=(), (output_dim,) or (n, output_dim)* -- Random process evaluated at
            the inputs.
        """
        raise NotImplementedError

    @property
    def input_dim(self) -> int:
        """Shape of inputs to the random process."""
        return self._input_dim

    @property
    def output_dim(self) -> int:
        """Shape of the random process evaluated at an input."""
        return self._output_dim

    @property
    def dtype(self) -> np.dtype:
        """Data type of (elements of) the random process evaluated at an input."""
        return self._dtype

    def marginal(self, args: _InputType) -> _randomvariablelist._RandomVariableList:
        """Batch of random variables defining the marginal distributions at the inputs.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Input(s) to evaluate random
            process at.
        """
        # return self.__call__(args).marginal()
        raise NotImplementedError

    @abc.abstractmethod
    def mean(self, args: _InputType) -> _OutputType:
        """Mean function.

        Returns the mean function evaluated at the given input(s).

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Input(s) where the mean
            function is evaluated.

        Returns
        -------
        _OutputType
            *shape=(), (output_dim, ) or (n, output_dim)* -- Mean function of the
            process evaluated at inputs ``x``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def cov(self, args0: _InputType, args1: Optional[_InputType] = None) -> _OutputType:
        r"""Covariance function or kernel.

        Returns the covariance function :math:`\operatorname{Cov}(f(x_0),
        f(x_1)) = \mathbb{E}[(f(x_0) - \mathbb{E}[f(x_0)])(f(x_0) - \mathbb{E}[f(
        x_0)])^\top]` of the process evaluated at :math:`x_0` and :math:`x_1`. If only
        ``args0`` is given the covariance among the components of the random process
        at the inputs defined by ``args0`` is computed.

        Parameters
        ----------
        args0
            *shape=(input_dim,) or (n0, input_dim)* -- First input to the covariance
            function.
        args1
            *shape=(input_dim,) or (n1, input_dim)* -- Second input to the covariance
            function.

        Returns
        -------
        _OutputType
            *shape=(), (output_dim, output_dim), (n0, n1) or (n0, n1, output_dim,
            output_dim)* -- Covariance of the process at ``args0`` and ``args1``. If
            only ``args0`` is given the kernel matrix :math:`K=k(x_0, x_0)` is computed.
        """  # pylint: disable=trailing-whitespace
        raise NotImplementedError

    def var(self, args: _InputType) -> _OutputType:
        """Variance function.

        Returns the variance function which is the value of the covariance or kernel
        evaluated elementwise at ``args`` for each output dimension separately.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Input(s) to the variance function.

        Returns
        -------
        _OutputType
            *shape=(), (output_dim,) or (n, output_dim)* -- Variance of the
            process at ``args``.
        """
        try:
            cov = self.cov(args0=args)
            if cov.ndim < 2:
                return cov
            elif cov.ndim == 2:
                return np.diag(cov)
            else:
                return np.vstack(
                    [np.diagonal(cov[:, :, i, i]) for i in range(self.output_dim)]
                ).T
        except NotImplementedError as exc:
            raise NotImplementedError from exc

    def std(self, args: _InputType) -> _OutputType:
        """Standard deviation function.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Input(s) to the standard
            deviation function.

        Returns
        -------
        _OutputType
            *shape=(), (output_dim,) or (n, output_dim)* -- Standard deviation of the
            process at ``args``.
        """
        try:
            return np.sqrt(self.var(args=args))
        except NotImplementedError as exc:
            raise NotImplementedError from exc

    @abc.abstractmethod
    def push_forward(
        self,
        args: _InputType,
        base_measure: Type[randvars.RandomVariable],
        sample: np.ndarray,
    ) -> np.ndarray:
        """Transform samples from a base measure into samples from the random process.

        This function can be used to control sampling from the random process by
        explicitly passing samples from a base measure evaluated at the input arguments.

        Parameters
        ----------
        args
            Input arguments.
        base_measure
            Base measure. Given as a type of random variable.
        sample
            *shape=(sample_size, output_dim)* -- Sample(s) from a base measure
            evaluated at the input arguments.
        """
        raise NotImplementedError

    def sample(
        self,
        args: _InputType = None,
        size: ShapeArgType = (),
        random_state: RandomStateArgType = None,
    ) -> Union[Callable[[_InputType], _OutputType], _OutputType]:
        """Sample paths from the random process.

        If no inputs are provided this function returns sample paths which are
        callables, otherwise random variables corresponding to the input locations
        are returned.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Evaluation input(s) of the
            sample paths of the process. If ``None``, sample paths, i.e. callables are
            returned.
        size
            Size of the sample.
        random_state :
            Random state of the random process. If None (or np.random), the global
            :mod:`numpy.random` state is used. If integer, it is used to seed the local
            :class:`~numpy.random.RandomState` instance.
        """
        if args is None:
            raise NotImplementedError

        return self._sample_at_input(args=args, size=size, random_state=random_state)

    @abc.abstractmethod
    def _sample_at_input(
        self,
        args: _InputType,
        size: ShapeArgType = (),
        random_state: RandomStateArgType = None,
    ) -> _OutputType:
        """Evaluate a set of sample paths at the given inputs.

        This function should be implemented by subclasses of :class:`RandomProcess`.
        This enables :meth:`sample` to both return functions, i.e. sample paths if
        only a `size` is provided and random variables if inputs are provided as well.

        Parameters
        ----------
        args
            *shape=(input_dim,) or (n, input_dim)* -- Evaluation input(s) of the
            sample paths of the process.
        size
            Size of the sample.
        random_state :
            Random state of the random process. If None (or np.random), the global
            :mod:`numpy.random` state is used. If integer, it is used to seed the local
            :class:`~numpy.random.RandomState` instance.
        """
        raise NotImplementedError



"""Tests for linear operators."""
import itertools
import unittest

import numpy as np

from probnum import linops
from tests.testing import NumpyAssertions


class LinearOperatorArithmeticTestCase(unittest.TestCase, NumpyAssertions):
    """Test linear operator arithmetic."""

    def setUp(self):
        """Resources for tests."""
        # Random Seed
        np.random.seed(42)

        # Scalars and arrays
        self.scalars = [0, int(1), 0.1, -4.2, np.nan, np.inf]
        self.arrays = [np.random.normal(size=[5, 4]), np.array([[3, 4], [1, 5]])]

    def test_scalar_mult(self):
        """Matrix linear operator multiplication with scalars."""
        for A, alpha in list(itertools.product(self.arrays, self.scalars)):
            with self.subTest():
                Aop = linops.Matrix(A)

                self.assertAllClose((alpha * Aop).todense(), alpha * A)

    def test_addition(self):
        """Linear operator addition."""
        for A, B in list(zip(self.arrays, self.arrays)):
            with self.subTest():
                Aop = linops.Matrix(A)
                Bop = linops.Matrix(B)

                self.assertAllClose((Aop + Bop).todense(), A + B)
